##
##  dev.sh -- OpenPKG Tool Chain
##  Copyright (c) 2000-2006 OpenPKG Foundation e.V. <http://openpkg.net/>
##  Copyright (c) 2000-2006 Ralf S. Engelschall <http://engelschall.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##
if [ ".$BASH" = . ]; then
    echo "openpkg:dev:ERROR: this tool must be run under bash" 1>&2
    exit 1
fi

#   decouple from the calling instance
unset OPENPKG_PREFIX

#   work around broken multiplexer in OpenPKG 2.0.0 and 2.0.1
#   this inhibits the caller from controlling additional and replacement tools through environment variables
if [ ".$OPENPKG_TOOLS_CMDSTACK" = . ]; then
    unset OPENPKG_TOOLS
    unset OPENPKG_TOOLS_APIPATH
    unset OPENPKG_TOOLS_CMDPATH
fi

# export variables
export OPENPKG_TOOLS
export OPENPKG_TOOLS_APIPATH
export OPENPKG_TOOLS_CMDPATH
export OPENPKG_TOOLS_EXEC
export OPENPKG_TOOLS_APIPATH_EXEC
export OPENPKG_TOOLS_CMDPATH_EXEC
export OPENPKG_TOOLS_TRUN
export OPENPKG_TOOLS_APIPATH_TRUN
export OPENPKG_TOOLS_CMDPATH_TRUN
export OPENPKG_ARGS

#   default configuration
OPENPKG_INST=`[ -r /etc/openpkg ] && grep /openpkg </etc/openpkg | sort -r; [ -r /etc/openpkg ] && grep -v /openpkg </etc/openpkg | sort -r`
OPENPKG_INST=`echo ${OPENPKG_INST:-/openpkg}`
OPENPKG_WORK=${HOME}/work/openpkg
OPENPKG_TEMP=${TMPDIR-/tmp}/${LOGNAME}/openpkg
#   special override for openpkg.net domain
case `hostname` in
*.openpkg.net )
    OPENPKG_MODE=developer
    OPENPKG_REPO=openpkg-cvs@cvs.openpkg.org:/v/openpkg/cvs
    OPENPKG_RMOD=openpkg-src
    OPENPKG_DIST=openpkg-ftp@ftp.openpkg.org:
    ;;
* )
    OPENPKG_MODE=contributor
    OPENPKG_REPO=:pserver:anonymous@cvs.openpkg.org:/v/openpkg/cvs
    OPENPKG_RMOD=openpkg-src
    OPENPKG_DIST=ftp://ftp.openpkg.org/contrib/00UPLOAD
    ;;
esac
export OPENPKG_INST OPENPKG_WORK OPENPKG_TEMP OPENPKG_MODE OPENPKG_REPO OPENPKG_RMOD OPENPKG_DIST

#   enforced and user controlled (through package/branch/execute) variables
OPENPKG_SPEC="${OPENPKG_SPEC}"; P="$P" # package, spec file without .spec suffix
OPENPKG_CTAG="${OPENPKG_CTAG}"; B="$B" # cvs tag, branch
OPENPKG_EXEC="${OPENPKG_EXEC}"; E="$E" # exec, instance
OPENPKG_TRUN="${OPENPKG_TRUN}"; T="$T" # trun, tools to run
OPENPKG_SDIR="${OPENPKG_SDIR}"; S="$S" # source dir
export OPENPKG_SPEC P OPENPKG_CTAG B OPENPKG_EXEC E OPENPKG_TRUN T OPENPKG_SDIR S

S="${S:-src}" # chicken-egg hack FIXME

#   use appropriate tools
EGREP=""
for i in `echo $PATH: | sed -e 's;:; ;g'` /usr/xpg4/bin; do
    i=${i}/egrep
    if [ -x $i -a .`echo foo | $i 2>/dev/null '^f[nop]+$'` = .foo ]; then
        EGREP=$i
        break;
    fi
done
AWK=""
for i in `echo $PATH: | sed -e 's;:; ;g'` /usr/xpg4/bin; do
    i=${i}/awk
    if [ -x $i -a .`echo foo | $i 2>/dev/null -v foo=bar '/foo/ { print foo }'` = .bar ]; then
        AWK=$i
        break;
    fi
done

#   a note about CVS HEAD
#   CVS calls the latest version of a file "head". It also supports two
#   hardcoded pseudotags, "BASE" and "HEAD". It is a strange and often
#   unexpected behaviour that checking out the "HEAD" tag is a sticky
#   operation and not identical to checking out the "head" and clear
#   stickyness which requires no -r option to "cvs checkout" or option
#   -A to "cvs up". In contrast to this original behaviour this program
#   uses "HEAD" for input, internal processing and output whenever the
#   user really means "clear stickyness". The "cd" shell script might
#   experience the unlikely event that a directory or file was really
#   updated to the "HEAD" tag. Only in this event the tag is being
#   output as "HEAD_STICKY" which means that this becomes a third
#   "hardcoded" pseudotag.

#   override configuration from ${HOME}/.openpkg-dev.rc (LEGACY)
if [ -f ${HOME}/.openpkg-dev.rc ]; then
    echo "openpkg:dev:WARNING: legacy ~/.openpkg-dev.rc found (please rename to ~/.openpkg/dev.rc)" 1>&2
    . ${HOME}/.openpkg-dev.rc
fi

#   override configuration from ${HOME}/.openpkg/dev.rc (OFFICIAL)
if [ -f ${HOME}/.openpkg/dev.rc ]; then
    . ${HOME}/.openpkg/dev.rc
fi

#   override configuration from command line
OPENPKG_ARGS=""
help=0
while [ $# -gt 0 ]; do
    opt=$1
    case ${opt} in
        -*=*) arg=`echo "${opt}" | sed 's/^[-_a-zA-Z0-9]*=//'` ;;
           *) arg='' ;;
    esac
    case ${opt} in
        -h|--help    ) help=1              ;;
        --inst=*     ) OPENPKG_INST=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--inst=${arg}";;
        --work=*     ) OPENPKG_WORK=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--work=${arg}";;
        --temp=*     ) OPENPKG_TEMP=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--temp=${arg}";;
        --mode=*     ) OPENPKG_MODE=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--mode=${arg}";;
        --repo=*     ) OPENPKG_REPO=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--repo=${arg}";;
        --rmod=*     ) OPENPKG_RMOD=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--rmod=${arg}";;
        --dist=*     ) OPENPKG_DIST=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--dist=${arg}";;
        --spec=*     ) OPENPKG_SPEC=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--spec=${arg}";;
        --ctag=*     ) OPENPKG_CTAG=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--ctag=${arg}";;
        --exec=*     ) OPENPKG_EXEC=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--exec=${arg}";;
        --trun=*     ) OPENPKG_TRUN=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--trun=${arg}";;
        --sdir=*     ) OPENPKG_SDIR=${arg} OPENPKG_ARGS="$OPENPKG_ARGS${OPENPKG_ARGS:+ }--sdir=${arg}";;
        -*           ) help="Invalid option \`${opt}'"; break ;;
        *            ) break               ;;
    esac
    shift
done

#   set P,B,E,T,S without having "cd()" already in place
[ ".$P" = . ] && P="${OPENPKG_SPEC}"
[ ".$B" = . ] && B="${OPENPKG_CTAG}"
[ ".$E" = . ] && E="${OPENPKG_EXEC}"
[ ".$T" = . ] && T="${OPENPKG_TRUN}"
[ ".$S" = . ] && S="${OPENPKG_SDIR}"

#   ------------------------------------------------------------

#   dump a function (optionally only the inner code part)
dumpfunc () {
    if [ ".$1" = ".-i" ]; then
        shift
        type $1 \
        | ${AWK} -v cmd=$1 'BEGIN { f=0 }
               /^ / { if (f==1) { print $0 } }
               /^[a-zA-Z0-9_]+ is a function/ { f=1 }'
    else
        type $1 \
        | ${AWK} -v cmd=$1 'BEGIN { f=0 }
               /.*/ { if (f==1) { print $0 } }
               /^[a-zA-Z0-9_]+ is a function/ { print "unalias "cmd" >/dev/null 2>&1"; f=1 }'
    fi
}

#   print a warning
warn () {
    echo "openpkg:dev:WARNING: $@" 1>&2
}

#   print an error
error () {
    echo "openpkg:dev:ERROR: $@" 1>&2
}

#   print an error and die
die () {
    error "$@"
    exit 1
}

makedir () {
    perm="$1"
    dir="$2"
    set -- `echo ":${dir}" | sed -e 's/^:\//%/' -e 's/^://' -e 's/\// /g' -e 's/^%/\//'`
    pathcomp=''
    for d in ${1+"$@"}; do
        pathcomp="${pathcomp}${d}"
        if [ ! -d "${pathcomp}" ]; then
            mkdir ${pathcomp} || errstatus=$?
            chmod ${perm} ${pathcomp} || errstatus=$?
        fi
        pathcomp="${pathcomp}/"
    done
}

#   provide a unified which(1) command, because vendor implementations differ in error results.
uwhich () {
    _rv=""
    for _tool in $*; do
        _ifs="$IFS"; IFS=":"
        for _path in $PATH; do
            if [ -f "$_path/$_tool" ]; then
                _rv="$_path/$_tool"
                break
            fi
        done
        IFS="$_ifs"; unset _ifs
        if [ ".$_rv" != . ]; then
            break
        fi
    done
    echo $_rv
    unset _rv
}

#   init MPX, RPM, OPENPKG_TOOLS_(API|CMD)PATH and RPMCMD
#   to the correct values for a given instance
#   see http://cvs.openpkg.org/chngview?cn=14911
initmpx ()
{
    if [ -x $1/bin/openpkg -a -x $1/libexec/openpkg/rpm ]; then
        #OpenPKG v2.x use "openpkg" multiplexer
        MPX="$1/bin/openpkg"
        RPM="rpm"
    elif [ -x $1/bin/rpm -a ! -x $1/libexec/openpkg/rpm ]; then
        #OpenPKG v1.x direct command execution
        MPX=""
        RPM="$1/bin/rpm"
    else
        # OpenPKG N/A
        MPX=""
        RPM=""
    fi
    RPMCMD="$MPX${MPX:+ }$RPM"
}

openpkgi () {
    if [ ".$1" = ".-e" ]; then
        if [ ".$E" != . ]; then
            initmpx $E
        else
            error "$0:ERROR: don't know which rpm to exec (E unset)" 1>&2
            return 1
        fi
        initmpx $E
        if [ ".$OPENPKG_TOOLS_EXEC" = . ];         then unset OPENPKG_TOOLS;         else OPENPKG_TOOLS="${OPENPKG_TOOLS_EXEC}";                 fi
        if [ ".$OPENPKG_TOOLS_APIPATH_EXEC" = . ]; then unset OPENPKG_TOOLS_APIPATH; else OPENPKG_TOOLS_APIPATH="${OPENPKG_TOOLS_APIPATH_EXEC}"; fi
        if [ ".$OPENPKG_TOOLS_CMDPATH_EXEC" = . ]; then unset OPENPKG_TOOLS_CMDPATH; else OPENPKG_TOOLS_CMDPATH="${OPENPKG_TOOLS_CMDPATH_EXEC}"; fi
    elif [ ".$1" = ".-t" ]; then
        if [ ".$T" != . ]; then
            initmpx $T
        else
            echo "$0:ERROR: don't know which tools to run (T unset)" 1>&2
            return 1
        fi
        initmpx $T
        if [ ".$OPENPKG_TOOLS_TRUN" = . ];         then unset OPENPKG_TOOLS;         else export OPENPKG_TOOLS="${OPENPKG_TOOLS_TRUN}";                 fi
        if [ ".$OPENPKG_TOOLS_APIPATH_TRUN" = . ]; then unset OPENPKG_TOOLS_APIPATH; else export OPENPKG_TOOLS_APIPATH="${OPENPKG_TOOLS_APIPATH_TRUN}"; fi
        if [ ".$OPENPKG_TOOLS_CMDPATH_TRUN" = . ]; then unset OPENPKG_TOOLS_CMDPATH; else export OPENPKG_TOOLS_CMDPATH="${OPENPKG_TOOLS_CMDPATH_TRUN}"; fi
    elif [ ".$1" = ".-i" ]; then
        shift
        initmpx $1
    else
        error "$0:ERROR: unknown option $1; USAGE: openpkgi [-e|-t|-i inst]" 1>&2
        return 1
    fi
    if [ ".${OPENPKG_TOOLS_CMDPROG}" = . ]; then
        error "OPENPKG_TOOLS_CMDPROG not set"
        return 1
    fi
    if [ ".${OPENPKG_TOOLS_CMDNAME}" = . ]; then
        error "OPENPKG_TOOLS_CMDNAME not set"
        return 1
    fi
    shift
    if [ ".$1" = .dev ]; then
        shift
        local cmd=""
        local arg=""
        local line=""
        local e;
        for e in "$@"; do
            if [ ".${cmd}" = . ]; then
                case $e in
                    -*) ;;
                    *) cmd="$e" ;;
                esac
            elif [ ".${arg}" = . ]; then
                case $e in
                    -*) ;;
                    *) arg="$e" ;;
                esac
            fi
        done
        case ${cmd} in
            package | pa* ) shift; openpkg_dev_package "${arg}"; cd . ;;
            branch  | br* ) shift; openpkg_dev_branch  "${arg}"; cd . ;;
            execute | ex* ) shift; openpkg_dev_execute "${arg}"; cd . ;;
            tools   | to* ) shift; openpkg_dev_tools   "${arg}"; cd . ;;
            srcdir  | sr* ) shift; openpkg_dev_srcdir  "${arg}"; cd . ;;
                        * ) command $T/bin/openpkg dev "$@" ;;
        esac
    elif [ ".$1" = .rpm ]; then
        shift
        for e in "$@"; do
            case $e in
                --rebuild) error "$0:ERROR: use of \"rpm --rebuild\" is disabled here to avoid destruction"; return 1 ;;
                --recompile) error "$0:ERROR: use of \"rpm --recompile\" is disabled here to avoid destruction"; return 1 ;;
            esac
        done
        if [ ".${RPMCMD}" = . ]; then
            error "$0:ERROR: no openpkg rpm found at instance \"$E\""
            return 1
        fi
        eval "HOME=\${OPENPKG_WORK} command ${RPMCMD} \"\$@\""
    else
        if [ ".${MPX}" = . ]; then
            error "$0:ERROR: no openpkg multiplexer found at instance \"$E\""
            return 1
        fi
        eval "HOME=\${OPENPKG_WORK} command ${MPX} \"\$@\""
    fi
}

openpkgE () {
    openpkgi -e "$@"
}

openpkgT () {
    openpkgi -t "$@"
}

openpkg () {
    if [ ".$1" = .dev ]; then
        openpkgi -t "$@"
    elif [ ".$1" = .rpm ]; then
        openpkgi -e "$@"
    elif [ ".$1" = .register ]; then
        openpkgi -e "$@"
    elif [ ".$1" = .rpm2cpio ]; then
        openpkgi -e "$@"
    elif [ ".$E" = ".$T" ]; then
	openpkgi -e "$@"
    else
        echo "$0:ERROR: don't know which openpkg multiplexer to call (unknown command $1)" 1>&2
        return 1
    fi
}

rpmi () {
    local mode
    mode=$1
    shift
    openpkgi $mode rpm "$@"
}

rpmE () {
    rpmi -e "$@"
}

rpmT () {
    rpmi -t "$@"
}

rpm () {
    openpkg rpm "$@"
}

rpm2cpioi () {
    local mode
    mode=$1
    shift
    openpkgi $mode rpm2cpio "$@"
}

rpm2cpioE () {
    rpm2cpioi -e "$@"
}

rpm2cpioT () {
    rpm2cpioi -t "$@"
}

rpm2cpio () {
    openpkg rpm2cpio "$@"
}

#   generate shell script for re-evaluation of command arguments
#   input:  foo 'bar  baz' quux       (3 args)
#   output: "foo \"bar  baz\" quux"   (1 arg)
_arg2sh () {
    local opt_e=no
    local opt_t=no
    local opt
    OPTIND=1
    while getopts et opt; do
        case ${opt} in
            e ) opt_e=yes ;;
            t ) opt_t=yes ;;
        esac
    done
    shift $(($OPTIND - 1))
    local cmd=""
    local arg
    for arg in "$@"; do
        #   NOTICE: there are three(!) escaping layers here:
        #   - escape layer 1: backticks (`..xx..`)
        #   - escape layer 2: quotes    ('..xx..')
        #   - escape layer 3: sed regex (s;..xx..;..xx..;)
        local orig=$arg
        if [ ${opt_e} = yes ]; then
            arg=`echo "${arg}" | sed -e 's;\\\\;\\\\\\\\;g' -e 's;\\(["]\\);\\\\\\1;g'`
        elif [ ${opt_t} = yes ]; then
            arg=`echo "${arg}" | sed -e 's;\\\\;\\\\\\\\;g' -e 's;\\(["!\$\`]\\);\\\\\\1;g'`
        fi
        local quote=no
        if [ "${arg}" != "${orig}" ]; then
            quote=yes
        else
            quote=`echo "X${arg}" | sed -e 's;^X.*[ 	].*\$;yes;' -e 's;^X.*;no;'`
        fi
        if [ ${quote} = yes ]; then
            arg="\"${arg}\""
        fi
        cmd="${cmd}${cmd:+ }${arg}"
    done
    printf "%s" "${cmd}" # intentionally not echo(1)
}

#   generate shell script for re-evaluation of command arguments
#   input:  foo 'bar  baz' quux       (3 args)
#   output: "foo \"bar  baz\" quux"   (1 arg)
_arg2sh () {
    local opt_e=no
    local opt_t=no
    local opt
    OPTIND=1
    while getopts et opt; do
        case ${opt} in
            e ) opt_e=yes ;;
            t ) opt_t=yes ;;
        esac
    done
    shift $(($OPTIND - 1))
    local cmd=""
    local arg
    for arg in "$@"; do
        #   NOTICE: there are three(!) escaping layers here:
        #   - escape layer 1: backticks (`..xx..`)
        #   - escape layer 2: quotes    ('..xx..')
        #   - escape layer 3: sed regex (s;..xx..;..xx..;)
        local orig=$arg
        if [ ${opt_e} = yes ]; then
            arg=`echo "@${arg}" | sed -e 's;^@;;' -e 's;\\\\;\\\\\\\\;g' -e 's;\\(["]\\);\\\\\\1;g'`
        elif [ ${opt_t} = yes ]; then
            arg=`echo "@${arg}" | sed -e 's;^@;;' -e 's;\\\\;\\\\\\\\;g' -e 's;\\(["!\$\`]\\);\\\\\\1;g'`
        fi
        local quote=no
        if [ "${arg}" != "${orig}" ]; then
            quote=yes
        else
            quote=`echo "@${arg}" | sed -e 's;^@.*[ 	].*\$;yes;' -e 's;^@.*;no;'`
        fi
        if [ ${quote} = yes ]; then
            arg=`echo "@${arg}" |\
                 sed -e 's;^@\([A-Z][A-Z0-9_]*\)=\(.*\)$;\1="\2";' \
                     -e 's;^@\(.*\)$;"\1";'`
        fi
        cmd="${cmd}${cmd:+ }${arg}"
    done
    printf "%s" "${cmd}" # intentionally not echo(1)
}

#   generate shell script for re-evaluation of environment variables
#   input:  PATH HOME                     (2 args)
#   output: "PATH=\"...\"; HOME=\"...\""  (1 arg)
_env2sh () {
    local cmd=""
    local arg
    for arg in "$@"; do
        local val
        eval "val=\"\$${arg}\""
        #   NOTICE: there are three(!) escaping layers here, too.
        val=`echo "@${val}" | sed -e 's;^@;;' -e 's;\\\\;\\\\\\\\;g' -e 's;\\(["!\$\`]\\);\\\\\\1;g'`
        cmd="${cmd}${cmd:+; }${arg}=\"${val}\""
    done
    printf "%s" "${cmd}" # intentionally not echo(1)
}

#   smart dealing with temporary root privileges
_root () {
    #   default operation
    if [ $# -eq 0 ]; then
        set -- -i
    fi

    #   parse command line options
    local opt_i=no # interactive shell
    local opt_l=no # last command-line
    local opt_e=no # remote expansion mode
    local opt
    OPTIND=1
    while getopts ileh opt; do
        case ${opt} in
            i ) opt_i=yes ;;
            l ) opt_l=yes ;;
            e ) opt_e=yes ;;
            h ) echo "root:Usage: root [-h] [-l] [-e] command"; return 0 ;;
            ? ) echo "root:Error: invalid command line";        return 1 ;;
        esac
    done
    shift $(($OPTIND - 1))

    #   determine action(s)
    local prolog=":"
    local epilog=":"
    local command
    if [ ${opt_i} = yes ]; then
        #   enter interactive session with root privileges
        local xtermcontrol=`uwhich xtermcontrol xtermcolors`
        if [ -n "${xtermcontrol}" ]; then
            prolog="${xtermcontrol} --cursor='#cc3333'"
            epilog="${xtermcontrol} --cursor='#000000'"
            command="PS1=\"\\[\\e[31;1m\\]\\u\\[\\e[0m\\]@\\h:\\w\\n\\\\\\$ \"; export PS1"
        fi
        command="${command}${command:+; }exec ${SHELL} --norc --noprofile"
    elif [ ${opt_l} = yes -a ${opt_e} = yes ]; then
        eval "set -- $(builtin history 2 | sed -n -e '1p' | cut -c8-)"
        command=$(_arg2sh -e -- "$@")
    elif [ ${opt_l} = yes ]; then
        #   execute last command-line with root privileges
        command=$(builtin history 2 | sed -n -e '1p' | cut -c8-)
    elif [ ${opt_e} = yes ]; then
        #   execute given command-line with root privileges (expansion mode)
        command=$(_arg2sh -e -- "$@")
    else
        #   execute given command-line with root privileges (transparent mode)
        command=$(_arg2sh -t -- "$@")
    fi

    #   generate action command
    local cmd=$(_env2sh ${ROOT_ENV:-PATH MANPATH INFOPATH LD_LIBRARY_PATH TERM HOME EDITOR PAGER LOGNAME USER})
    cmd="${cmd}; cd ${PWD} && ${command}"
    cmd=$(_arg2sh -t -- "${cmd}")
    eval ${prolog}; ssh -t -q -x root@${HOSTNAME} ${SHELL} -c "${cmd}"; rv=$?; eval ${epilog}

    #   exit with return value of remote command
    return $rv
}

#   smart dealing with temporary root privileges with dev.sh specifics
root () {
    if [ ".$1" = .rpm ]; then
        initmpx $E
        shift
        if [ ".${RPMCMD}" = . ]; then
            echo "$0:ERROR: no openpkg rpm found at instance \"$E\""
            return 1
        fi
        _root HOME=${OPENPKG_WORK} command ${RPMCMD} "$@"
    elif [ ".$1" = .openpkg ]; then
        initmpx $E
        shift
        if [ ".${MPX}" = . ]; then
            echo "$0:ERROR: no openpkg multiplexer found at instance \"$E\""
            return 1
        fi
        _root HOME=${OPENPKG_WORK} command ${MPX} "$@"
    else
        _root "$@"
    fi
}

cvs () {
    cvs_colorize="cat"
    case "$1" in
        di|dif|diff   ) [ ".`uwhich cvs-diff-colorize`"   != . ] && cvs_colorize="cvs-diff-colorize"   ;;
        up|upd|update ) [ ".`uwhich cvs-update-colorize`" != . ] && cvs_colorize="cvs-update-colorize" ;;
        com|commit    ) error "$0:ERROR: use of \"cvs commit\" is disabled here, use \"opd release\""; return 1 ;;
        ci|checkin    ) error "$0:ERROR: use of \"cvs checkin\" is disabled here, use \"opd release\""; return 1 ;;
    esac
    HOME=${OPENPKG_WORK} command cvs ${1+"$@"} | $cvs_colorize
}

#   shared code between "build" and "release"
_autorel () {
    echo "++ determine old version (if any)"
    tmpfile="${OPENPKG_TEMP}/${name}${GUID}"
    cvs_stat=`builtin cd ${OPENPKG_WORK}; cvs -d ${OPENPKG_REPO} stat $S/${name}/${name}.spec 2>/dev/null | grep 'Repository revision:'`
    cvs_rev=`echo "${cvs_stat}" | ${AWK} '/Repository revision:/ { print $3; }'`
    cvs_dead=`echo "${cvs_stat}" | grep '/Attic/'`
    if [ ".${cvs_dead}" != . ]; then
        cvs_rev="No"
    fi
    if [ ".${cvs_rev}" != .No ]; then
        cvs -d ${OPENPKG_REPO} co -p -r${cvs_rev} "${OPENPKG_RMOD}/${name}/${name}.spec" >${tmpfile} 2>/dev/null
        rpm_version_old=`rpmE -q --qf "%{VERSION}" --specfile ${tmpfile} 2>/dev/null`
        rpm_release_old=`rpmE -q --qf "%{RELEASE}" --specfile ${tmpfile} 2>/dev/null`
        command rm -f ${tmpfile}
        [ ".${rpm_version_old}" = . ] && die "cannot determine old version number"
        [ ".${rpm_release_old}" = . ] && die "cannot determine old release number"
    fi

    echo "++ determine new version"
    rpm_version_new=`rpmE -q --qf "%{VERSION}" --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
    rpm_name=`rpmE -q --qf "%{NAME}" --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
    rpm_summary=`rpmE -q --qf "%{SUMMARY}" --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
    [ ".${rpm_version_new}" = . ] && die "cannot determine new version number"

    echo "++ determine branch and related release number and upload suffix"
    rpm_release_new=`rpmE -q --qf "%{RELEASE}" --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
    [ ".${rpm_release_new}" = . ] && die "cannot determine new release number"

    if [ ".$B" = .HEAD ]; then
        BRANCH="HEAD"
    else
        BRANCH=`echo "$B" | sed -e 's;^OPENPKG_;;' -e 's;_STABLE$;;' -e 's;SOLID$;;' -e 's;^HEAD$;;' -e 's;_;;g'`
        OPMAJORV=`echo ${BRANCH} | cut -c 1`
        OPMINORV=`echo ${BRANCH} | cut -c 2`
        if [ ".${OPMINORV}" != . ]; then
            BRANCH="SOLID"
        elif [ ".${OPMAJORV}" != . ]; then
            BRANCH="STABLE"
        else
            BRANCH=""
        fi
    fi
    rpm_release_req=""
    upload=""
    rpmwipe=""
    if [ ".${BRANCH}" = .HEAD ]; then
        rpm_release_req=`date '+%Y%m%d'`
        upload="current/SRC"
    fi
    if [ ".${BRANCH}" = .STABLE ]; then
        rpm_release_req="${OPMAJORV}.`date '+%Y%m%d'`"
        upload="stable/${OPMAJORV}/SRC"
    fi
    if [ ".${BRANCH}" = .SOLID ]; then
        update=`echo ${rpm_release_old} | sed -e 's;^.*\.;;'`
        if [ ".${update}" = .0 ]; then
            RELEASE=`echo "$B" | sed -e 's;_SOLID$;_RELEASE;'`
            cvs_release=`builtin cd ${OPENPKG_WORK}; cvs -d ${OPENPKG_REPO} log -r${cvs_rev} $S/${name}/${name}.spec 2>/dev/null | fgrep $RELEASE`
            if [ ".${cvs_release}" = . ]; then
                echo "++ detected post-branch pre-release engineering phase - keep zero update number"
                rpm_release_req="${OPMAJORV}.${OPMINORV}.0"
                upload="release/${OPMAJORV}.${OPMINORV}/SRC"
                rpmwipe="${name}-${rpm_version_old}-${rpm_release_old}"
                [ ".${rpmfile}" = ".${rpmwipe}" ] && rpmwipe=""
            else
                update=`expr ${update} + 1`
                rpm_release_req="${OPMAJORV}.${OPMINORV}.${update}"
                upload="release/${OPMAJORV}.${OPMINORV}/UPD/00UPLOAD"
            fi
        else
            update=`expr ${update} + 1`
            rpm_release_req="${OPMAJORV}.${OPMINORV}.${update}"
            upload="release/${OPMAJORV}.${OPMINORV}/UPD/00UPLOAD"
        fi
    fi
}

setup () {
    #   honor optional parameters [[[spec]ctag]exec]; use "" to specify a empty parameter and use it's default
    [ ".$1" != . ] && OPENPKG_SPEC="$1"
    [ ".$2" != . ] && OPENPKG_CTAG="$2"
    [ ".$3" != . ] && OPENPKG_EXEC="$3"

    #   OPENPKG_SPEC in setup() is only a flag to indicate minimal checkout

    #   canonify OPENPKG_CTAG, default to HEAD
    canonifybranch "${OPENPKG_CTAG}"
    OPENPKG_CTAG="${RV:-HEAD}"

    #   canonify OPENPKG_EXEC, default to HEAD
    canonifybranch "${OPENPKG_EXEC:-$OPENPKG_CTAG}"
    matchinstance "${RV-HEAD}"
    OPENPKG_EXEC="${RV}"

    #   sanity check
    case "${OPENPKG_WORK}" in
        /?* ) ;;
          * ) die "working area ${OPENPKG_WORK} seems to be strange" ;;
    esac
    if [ -d "${OPENPKG_WORK}" ]; then
        die "working area ${OPENPKG_WORK} already exists"
    fi

    #   create a working area
    echo "++ creating working area ${OPENPKG_WORK}"
    if [ ! -d "${OPENPKG_WORK}" ]; then
        makedir 755 "${OPENPKG_WORK}" || die "failed to create ${OPENPKG_WORK} directory"
    fi

    #   sane environment
    umask 022

    #   create a .cvspass
    echo "++ creating CVS password file (${OPENPKG_WORK}/.cvspass)"
    sed -e 's;^ *;;' <<EOF >"${OPENPKG_WORK}/.cvspass"
        /1 ${OPENPKG_REPO} A
EOF

    #   create a .cvsrc
    echo "++ creating CVS configuration (${OPENPKG_WORK}/.cvsrc)"
    updatedir=""; [ ".${P}" = . ] && updatedir="-d"
    sed -e 's;^ *;;' <<EOF >"${OPENPKG_WORK}/.cvsrc"
        checkout -P
        update -P $updatedir
        diff -u -d
        rdiff -u
EOF

    echo "++ creating CVS file ignore list (${OPENPKG_WORK}/.cvsignore)"
    sed -e 's;^ *;;' <<EOF >"${OPENPKG_WORK}/.cvsignore"
        .[a-za-Z0-9]*
        00README
        dst
        pkg
EOF

    #   create a .rpmmacros, save existing, overwrite only if different
    echo "++ creating RPM configuration (${OPENPKG_WORK}/.rpmmacros)"
    sed -e 's;^ *;;' <<EOF >"${OPENPKG_WORK}/.rpmmacros"
        #   source areas
        %_sourcedir    %(echo \${OPENPKG_WORK})/dst/%{name}
        %_specdir      %(echo \${OPENPKG_WORK})/%(echo \$S)/%{name}

        #   temporary areas
        %_builddir     %(echo \${OPENPKG_TEMP})
        %_tmppath      %(echo \${OPENPKG_TEMP})

        #   target areas
        %_rpmdir       %(echo \${OPENPKG_WORK})/pkg/bin
        %_srcrpmdir    %(echo \${OPENPKG_WORK})/pkg/src
EOF

    echo "++ creating directory structure (${OPENPKG_REPO}/{dst,pkg})"
    #   special override for openpkg.net domain
    case `hostname` in
        *.openpkg.net )
            ( echo ""
              echo "#   ==== Overide settings for C&W IS Development Team ==="
              echo ""
              echo "#   use local mirror first"
              echo "%l_fetch_mirror_0 ftp://ftp.openpkg.org/sources/DST/%{name}/"
              echo ""
              echo "#   use shared storage"
              echo "%_sourcedir       /v/openpkg/DST/%{name}"
              echo "%_rpmdir          /v/openpkg/PKG/bin"
              echo "%_srcrpmdir       /v/openpkg/PKG/src"
            ) >>${OPENPKG_WORK}/.rpmmacros
            if [ ! -d "${OPENPKG_WORK}/dst" ]; then
                ln -s /v/openpkg/DST "${OPENPKG_WORK}/dst" || die "failed to link to ${OPENPKG_WORK}/dst directory"
            fi
            if [ ! -d "${OPENPKG_WORK}/pkg" ]; then
                ln -s /v/openpkg/PKG "${OPENPKG_WORK}/pkg" || die "failed to link to ${OPENPKG_WORK}/pkg directory"
            fi
            ;;
        * )
            if [ ! -d "${OPENPKG_WORK}/dst" ]; then
                makedir 755 "${OPENPKG_WORK}/dst" || die "failed to create ${OPENPKG_WORK}/dst directory"
            fi
            if [ ! -d "${OPENPKG_WORK}/pkg/src" ]; then
                makedir 755 ${OPENPKG_WORK}/pkg/src || die "failed to create ${OPENPKG_WORK}/pkg/src directory"
            fi
            if [ ! -d "${OPENPKG_WORK}/pkg/bin" ]; then
                makedir 755 ${OPENPKG_WORK}/pkg/bin || die "failed to create ${OPENPKG_WORK}/pkg/bin directory"
            fi
    esac

    #   checkout the CVS module(s)
    if [ ".${P}" = . ]; then
        echo "++ checking out source areas (${OPENPKG_REPO}/{adm,doc,re,src,web,registry,tools})"
        ( builtin cd "${OPENPKG_WORK}"
          HOME="${OPENPKG_WORK}"
          export HOME
          cvs -q -d "${OPENPKG_REPO}" checkout -d adm      openpkg-adm
          cvs -q -d "${OPENPKG_REPO}" checkout -d doc      openpkg-doc
          cvs -q -d "${OPENPKG_REPO}" checkout -d re       openpkg-re
          cvs -q -d "${OPENPKG_REPO}" checkout -d web      openpkg-web
          cvs -q -d "${OPENPKG_REPO}" checkout -d tools    openpkg-tools
          cvs -q -d "${OPENPKG_REPO}" checkout -d registry openpkg-registry
          if [ ".${B:-HEAD}" = .HEAD ]; then
              cvs -q -d "${OPENPKG_REPO}" checkout -d src "${OPENPKG_RMOD}"
          else
              cvs -q -d "${OPENPKG_REPO}" checkout -r ${B} -d src "${OPENPKG_RMOD}"
          fi
        )
    else
        echo "++ checking out minimal source areas (${OPENPKG_REPO}/{re,src,tools})"
        ( builtin cd "${OPENPKG_WORK}"
          HOME="${OPENPKG_WORK}"
          export HOME
          cvs -q -d "${OPENPKG_REPO}" checkout -d re    openpkg-re
          cvs -q -d "${OPENPKG_REPO}" checkout -d tools openpkg-tools
          if [ ".${B:-HEAD}" = .HEAD ]; then
              cvs -q -d "${OPENPKG_REPO}" checkout -d src "${OPENPKG_RMOD}/00README"
              cvs -q -d "${OPENPKG_REPO}" checkout -d src "${OPENPKG_RMOD}/${P}"
          else
              cvs -q -d "${OPENPKG_REPO}" checkout -r ${B} -d src "${OPENPKG_RMOD}/00README"
              cvs -q -d "${OPENPKG_REPO}" checkout -r ${B} -d src "${OPENPKG_RMOD}/${P}"
          fi
        )
    fi

    echo "++ creating README file (${OPENPKG_REPO}/00README)"
    help

    #   cvs with RSE patches reads ~/.cvsrc, the .cvsrc file in the
    #   current directory and all directories towards / and merges
    #   all the options it finds.
    #   FIXME
    #   - patch the RSE patch to disable the upward search -> bad for existing binaries
    #   - use -f option to ignore .cvsrc -> requires this script to provide the options directly
    #   - temporary solution is to just print a warning
    cvs --version | grep RSE >/dev/null && if [ -f $HOME/.cvsrc ]; then
        warn "you use cvs with RSE patches, your ~/.cvsrc may collide with ${OPENPKG_WORK}/.cvsrc"
    fi
}

update () {
    full=0
    join=0
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-f ) full=1 ;;
            .-j ) join=1 ;;
              * ) packages="${packages} $1"
        esac
        shift
    done
    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            full=1
        fi
        all="$P"
    else
        all="${packages}"
    fi
    if [ $full -eq 0 ]; then
        for name in $all; do
            [ -d ${OPENPKG_WORK}/$S/${name}/CVS ] || continue;
            if [ ${join} -eq 0 ]; then
                echo "++ updating package ${name}"
                (builtin cd ${OPENPKG_WORK}/$S/${name} && cvs update)
            else
                echo "++ joining package ${name} with HEAD"
                (builtin cd ${OPENPKG_WORK}/$S/${name} || exit 1
                 tmpfile="${OPENPKG_TEMP}/${name}${GUID}"
                 haverelease=`${EGREP} <${name}.spec '^Release:'`
                 havebprereq=`${EGREP} <${name}.spec '^BuildPreReq: .*openpkg *>= *' | sed -e 's;^.*openpkg *>= *;;' -e 's;,* .*$;;'`
                 haverprereq=`${EGREP} <${name}.spec      '^PreReq: .*openpkg *>= *' | sed -e 's;^.*openpkg *>= *;;' -e 's;,* .*$;;'`
                 cvs -d ${OPENPKG_REPO} co -p -r1 "${OPENPKG_RMOD}/${name}/${name}.spec" >${tmpfile} 2>/dev/null
                 joinrelease=`${EGREP} <${tmpfile} '^Release: '`
                 joinversion=`${EGREP} <${tmpfile} '^Version: '`
                 joinbprereq=`${EGREP} <${tmpfile} '^BuildPreReq: .*openpkg *>= *' | sed -e 's;^.*openpkg *>= *;;' -e 's;,* .*$;;'`
                 joinrprereq=`${EGREP} <${tmpfile}      '^PreReq: .*openpkg *>= *' | sed -e 's;^.*openpkg *>= *;;' -e 's;,* .*$;;'`
                 sed <${name}.spec >${name}.spec.n \
                    -e "s;^Release: .*$;$joinrelease;" \
                    -e "s;^Version: .*$;$joinversion;" \
                    -e "s;^\(BuildPreReq: .*openpkg *>= *\)[^,]*\(.*\)$;\1$joinbprereq\2;" \
                    -e      "s;^\(PreReq: .*openpkg *>= *\)[^,]*\(.*\)$;\1$joinrprereq\2;" \
                    && mv ${name}.spec.n ${name}.spec
                 cvs update -j1
                 sed <${name}.spec >${name}.spec.n \
                    -e "s;^Release: .*;$haverelease;" \
                    -e "s;^\(BuildPreReq: .*openpkg *>= *\)[^,]*\(.*\)$;\1$havebprereq\2;" \
                    -e      "s;^\(PreReq: .*openpkg *>= *\)[^,]*\(.*\)$;\1$haverprereq\2;" \
                    && mv ${name}.spec.n ${name}.spec
                 command rm ${tmpfile}
                )
            fi
        done
    else
        echo "++ updating ${OPENPKG_REPO}/{adm,doc,re,src.*,web,tools}"
        ( builtin cd ${OPENPKG_WORK} || die "cannot cd to ${OPENPKG_WORK}"
          HOME=${OPENPKG_WORK}
          export HOME
          for subdir in adm doc re src src.* web tools; do
              if [ -d $subdir ]; then
                  cvs -d ${OPENPKG_REPO} update $subdir
              fi
          done
        )
    fi
}

package () {
    warn "running \"package\" is useless until called from within interactive development shell"
}

openpkg_dev_package () {
    local package=""
    if [ ".$1" = .-q -o ".$1" = .--query ]; then
        package=`echo ${PWD} | sed -e "s;^${OPENPKG_WORK}/$S/;;" -e 's;/.*$;;'`
        echo "${package}"
    else
        package="$1"
        if [ ".${package}" != . ]; then
            if [ ! -d "${OPENPKG_WORK}/$S" ]; then
                error "${OPENPKG_WORK}/$S directory not found"
                return 1
            fi

            ( builtin cd ${OPENPKG_WORK}/$S && HOME=${OPENPKG_WORK} cvs up -d ${package} >/dev/null 2>&1 )
            if [ ! -d "${OPENPKG_WORK}/$S/${package}" ]; then
                error "${OPENPKG_WORK}/$S/${package} directory not found"
                return 1
            fi
            builtin cd "${OPENPKG_WORK}/$S/${package}"
        fi
        OPENPKG_SPEC="${package}"
        cd .
    fi
}

branch () {
    warn "running \"branch\" is useless until called from within interactive development shell"
}

openpkg_dev_branch () {
    local branch=""
    if [ ".$1" = .-q -o ".$1" = .--query ]; then
        cvstagfiles=""
        if [ -r CVS/Entries ]; then
            cvstagfiles="${cvstagfiles} CVS/Entries"
        fi
        if [ -r CVS/Tag ]; then
            cvstagfiles="${cvstagfiles} CVS/Tag"
        fi
        if [ ".${cvstagfiles}" != . ]; then
            branch=`cat ${cvstagfiles} \
            | sed -e 's;//THEAD$;//THEAD_STICKY;' -e 's;//$;//THEAD;' -n -e 's;^.*//T;;p' \
            | sort | uniq | tr '\n' ':' | sed -e 's;:$;;'`
        fi
        echo "${branch}"
    else
        branch="$1"
        if [ ".${branch}" != . ]; then
            if [ ".${branch}" != .-A -a ".${OPENPKG_TOOLS_CMDPROG}" != . -a ".${OPENPKG_TOOLS_CMDNAME}" != . ]; then
                branch="`${OPENPKG_TOOLS_CMDPROG} ${OPENPKG_TOOLS_CMDNAME} query canonifybranch \"${branch}\"`"
            fi
            if [ ".${branch}" = .HEAD ]; then
                HOME=${OPENPKG_WORK} cvs update -A
            elif [ ".${branch}" = .-A ]; then
                HOME=${OPENPKG_WORK} cvs update -A
                branch=""
            else
                HOME=${OPENPKG_WORK} cvs update -r ${branch}
            fi
        fi
        OPENPKG_CTAG=${branch}
        cd .
    fi
}

execute () {
    warn "running \"execute\" is useless until called from within interactive development shell"
}

openpkg_dev_execute () {
    local execute=""
    if [ ".$1" = .-q -o ".$1" = .--query ]; then
        #FIXME matchinstance "${B}"
        execute="$E"
        echo "${execute}"
    else
        execute="$1"
        if [ ".${execute}" != . ]; then
            initmpx ${execute}
            if [ ".${RPMCMD}" = . ]; then
                error "${execute} is not a OpenPKG instance"
                return 1
            fi
        fi
        OPENPKG_EXEC="${execute}"
        cd .
    fi
}

tools () {
    warn "running \"tools\" is useless until called from within interactive development shell"
}

openpkg_dev_tools () {
    local tools=""
    if [ ".$1" = .-q -o ".$1" = .--query ]; then
        tools="$E"
        echo "${tools}"
    else
        tools="$1"
        if [ ".${tools}" != . ]; then
            initmpx ${tools}
            if [ ".${RPMCMD}" = . ]; then
                error "${tools} is not a OpenPKG instance"
                return 1
            fi
        fi
        OPENPKG_TRUN="${tools}"
        cd .
    fi
}

srcdir () {
    warn "running \"srcdir\" is useless until called from within interactive development shell"
}

openpkg_dev_srcdir () {
    local srcdir=""
    if [ ".$1" = .-q -o ".$1" = .--query ]; then
        srcdir=`pwd | sed -e 's;^;X;' -e 's;^X.*/\(src\.[^/][^/]*\).*$;\1;' -e 's;^X.*;src;'`
        echo "${srcdir}"
    else
        srcdir="$1"
        if [ ".${srcdir}" != . ]; then
            if [ ! -d "${OPENPKG_WORK}/$1" ]; then
                error "${srcdir} is not a directory"
                return 1
            fi
            OPENPKG_SDIR="$1"
            cd
        else
            OPENPKG_SDIR="${srcdir}"
            cd .
        fi
    fi
}

#   convert any (branch|instance|name) into branch and return ${RV}
#
#   Branch              Instance   Name
#   ------------------- ---------- --------
#   HEAD                /cw        CURRENT
#   OPENPKG_1_STABLE    /cw1       1-STABLE
#   OPENPKG_1_2_SOLID   /cw12      1.2
#   OPENPKG_1_1_SOLID   /cw11      1.1
#   OPENPKG_1_0_SOLID   /cw10      1.0
#
canonifybranch () {
    #   remove surrounding whitespaces, take first arg only, toupper
    RV=`echo ${1} | cut -f 1 -d" " | tr '[a-z]' '[A-Z]'`

    #   empty in, empty return (mapping emtpy to HEAD is the task of the caller)
    [ ".${RV}" = . ] && return

    #   strip off any abbreviated form of a leading OPENPKG_
    RV=`echo "${RV}" | sed -e 's;^O\{0,1\}P\{0,1\}E\{0,1\}N\{0,1\}P\{0,1\}K\{0,1\}G\{0,1\}_\{0,1\};;'`

    #   transform instance into branch
    RV=`echo "${RV}" | sed -e 's;^/[A-Z]\{1,\}$;HEAD;' -e 's;^/[A-Z]\{1,\}\([1-9]\);\1_;'`

    #   replace dots from (SOLID name) and dashes (STABLE name) with underscore for convenience
    RV=`echo "${RV}" | sed -e 's;[\.-];_;g'`

    #   invalid branch, empty return (mapping emtpy to HEAD is the task of the caller)
    [ ".${RV}" = . ] && return

    #   check for HEAD (branch) aka CURRENT (name)
    echo "CURRENT" | ${EGREP} "^${RV}" >/dev/null 2>&1 && RV="HEAD"
    echo "HEAD"    | ${EGREP} "^${RV}" >/dev/null 2>&1 && RV="HEAD"

    if [ .${RV} != ".HEAD" ]; then

        #   extract branchpoint (1, 1_1 ...) and optional direction
        BRANBP=`echo "${RV}" | sed -e 's;^\([1-9][0-9]*\(_\{0,1\}[0-9]*\)\)\(.*\)$;\1;' -e 's;_*$;;'`
        BRANDR=`echo "${RV}" | sed -e 's;^\([1-9][0-9]*\(_\{0,1\}[0-9]*\)\)\(.*\)$;\3;' -e 's;^_*;;'`

        #   select direction, check optional direction
        echo ${BRANBP} | grep '_' >/dev/null
        if [ $? -eq 0 ]; then
            RV="OPENPKG_${BRANBP}_SOLID"
        else
            RV="OPENPKG_${BRANBP}_STABLE"
        fi

        unset BRANBP
        unset BRANDR
    fi
}

#   find matching instance with a executeable rpm for a given branch $1; return "" if none found
#   $1 = requested branch
#   $2 = optional mode one of "exact", "best", "any"; if omitted, defaults to "any"
#        exact = find executable rpm exactly matching branch
#        any   = take first executable rpm
matchinstance () {
    mode="${2:-any}"
    [ ".$mode" != .exact -a ".$mode" != .any ] && die "matchinstance() assertion"

    INSTANCE="$1"
    if [ ".$mode" = .exact ]; then
        echo "${INSTANCE}" | ${EGREP} "^/" >/dev/null #PATH
        if [ $? -eq 0 ]; then
            initmpx ${INSTANCE}
            if [ ".${RPMCMD}" != . ]; then
                RV="${INSTANCE}"
                unset INSTANCE
                return
            else
                RV=""
                unset INSTANCE
                return
            fi
        fi
    fi

    INSTANCE=`echo "$1" | sed -e 's;^OPENPKG_;;' -e 's;_STABLE$;;' -e 's;_SOLID$;;' -e 's;^HEAD$;;' -e 's;_;.;g'`
    RE=""
    echo "${INSTANCE}" | ${EGREP} -i "^head$"                    >/dev/null && RE="^[0-9]{8}$"     #HEAD (any)
    echo "${INSTANCE}" | ${EGREP} "^[0-9]{8}$"                   >/dev/null && RE="^${INSTANCE}$"  #HEAD
    echo "${INSTANCE}" | ${EGREP} "^[0-9]{1,7}$"                 >/dev/null && RE="^${INSTANCE}\." #STABLE
    echo "${INSTANCE}" | ${EGREP} "^[0-9]+\.[0-9]{8}$"           >/dev/null && RE="^${INSTANCE}$"  #SNAPSHOT
    echo "${INSTANCE}" | ${EGREP} "^[0-9]+\.[0-9]{1,7}$"         >/dev/null && RE="^${INSTANCE}"   #SOLID (any UPD)
    echo "${INSTANCE}" | ${EGREP} "^[0-9]+\.[0-9]{1,7}\.[0-9]+$" >/dev/null && RE="^${INSTANCE}$"  #SOLID
    RE="`echo ${RE} | sed -e 's;\.;\\.;g'`"

    if [ ".$RE" != . ]; then
        for INSTANCE in ${OPENPKG_INST}; do
            initmpx ${INSTANCE}
            if [ ".${RPMCMD}" != . ]; then
                VERSION=`${RPMCMD} -q --qf "%{VERSION}" openpkg`
                echo ${VERSION} | ${EGREP} "$RE" >/dev/null
                if [ $? -eq 0 ]; then
                    RV="${INSTANCE}"
                    unset INSTANCE
                    return
                fi
            fi
        done
    fi

    if [ ".$mode" = .any ]; then
        OPENPKG_INST="${OPENPKG_INST} `uwhich openpkg | sed -e 's;/bin/openpkg$;;'`"
        OPENPKG_INST="${OPENPKG_INST} `uwhich rpm | sed -e 's;/bin/rpm$;;'`"
        for INSTANCE in ${OPENPKG_INST}; do
            initmpx ${INSTANCE}
            if [ ".${RPMCMD}" != . ]; then
                RV="${INSTANCE}"
                unset INSTANCE
                return
            fi
        done
    fi
    RV=""
    unset INSTANCE
}

bashrcprolog () {
    #   use original .bashrc as a baseline
    [ -f ${HOME}/.bashrc ] && source ${HOME}/.bashrc

    #   use SSH for CVS
    CVS_RSH=${CVS_RSH:-ssh}
    export CVS_RSH

    #   import OpenPKG environment
    #eval `$E/etc/rc --eval openpkg env` #FIXME no longer needed (i.e. MANPATH); anyway: wrong place here!

    #   prompt to indicate openpkg dev bash environment
    declare -a DIRSTACKREV=()
    cd () {
        local PX PD BX BD EX ED TX TD SX SD IX ID workdir
        if [ ".$1" = .- ]; then
            #   go to previous working directory on forward directory stack
            #   and move this directory onto the reverse directory stack
            if [ ${#DIRSTACK[*]} -gt 1 ]; then
                workdir="${DIRSTACK[0]}"
                DIRSTACKREV[${#DIRSTACKREV[*]}]="${workdir}"
                popd >/dev/null
            else
                echo "cd: no more previous working directories on forward directory stack" 1>&2
                return 1
            fi
        elif [ ".$1" = .+ ]; then
            #   go to previous working directory on reverse directory stack
            #   and move this directory onto the forward directory stack
            if [ ${#DIRSTACKREV[*]} -gt 0 ]; then
                local i=$((${#DIRSTACKREV[*]} - 1))
                eval "pushd ${DIRSTACKREV[$i]} >/dev/null"
                unset DIRSTACKREV[$i]
            else
                echo "cd: no more previous working directories on reverse directory stack" 1>&2
                return 1
            fi
        else
            #   go to next working directory
            #   and move this directory onto the forward directory stack
            if [ ".$1" = . ]; then
                if [ ".${OPENPKG_SDIR}" = . ]; then
                    workdir="${OPENPKG_WORK}/src/${OPENPKG_SPEC:+/}${OPENPKG_SPEC}"
                else
                    workdir="${OPENPKG_WORK}/${OPENPKG_SDIR}/${OPENPKG_SPEC:+/}${OPENPKG_SPEC}"
                fi
            else
                workdir="$1"
            fi
            if [ -d "${workdir}" ]; then
                pushd ${workdir} >/dev/null
                if [ ${#DIRSTACK[*]} -ge 2 -a ${DIRSTACK[0]} = ${DIRSTACK[1]} ]; then
                    popd >/dev/null
                fi
            fi
            DIRSTACKREV=()
        fi

        P=`openpkg_dev_package --query`
        # P holds detection
        if [ ".${OPENPKG_SPEC}" = . ]; then
            if [ ".${P}" = . ]; then
                # nothing forced, nothing detected; keep nothing; print nothing
                :
            else
                # nothing forced, something detected; keep detected; print detected normal
                PX="P=${P}"
                PD="P=${P}"
            fi
        else
            if [ ".${P}" = . ]; then
                # forced, nothing detected; take forced; print forced black (aka trailing dot)
                P="${OPENPKG_SPEC}"
                PX="[1mP=${P}[m"
                PD="P=${P}."
            else
                if [ ".${OPENPKG_SPEC}" = ".${P}" ]; then
                    # forced; detected and matched; keep; print black (aka trailing dot)
                    PX="[1mP=${P}[m"
                    PD="P=${P}."
                else
                    # forced; detected and mismatch; keep forced; print forced red (aka trailing exclamation mark)
                    PX="[31mP=${OPENPKG_SPEC}[m"
                    PD="P=${OPENPKG_SPEC}!"
                    P="${OPENPKG_SPEC}"
                fi
            fi
        fi
        # P has ultimate content

        B=`openpkg_dev_branch --query`
        # B holds detection
        if [ ".${OPENPKG_CTAG}" = . ]; then
            if [ ".$B" = . ]; then
                # nothing forced, nothing detected; keep nothing; print nothing
                :
            else
                # nothing forced, something detected; keep detected; print detected normal
                BX="B=$B"
                BD="B=$B"
            fi
        else
            if [ ".$B" = . ]; then
                # forced, nothing detected; take forced; print forced black (aka trailing dot)
                B="${OPENPKG_CTAG}"
                BX="[1mB=$B[m"
                BD="B=$B."
            else
                if [ ".${OPENPKG_CTAG}" = ".$B" ]; then
                    # forced; detected and matched; keep; print black (aka trailing dot)
                    BX="[1mB=$B[m"
                    BD="B=$B."
                else
                    # forced; detected and mismatch; keep forced; print forced red (aka trailing exclamation mark)
                    BX="[31mB=${OPENPKG_CTAG}[m"
                    BD="B=${OPENPKG_CTAG}!"
                    B="${OPENPKG_CTAG}"
                fi
            fi
        fi
        # B has ultimate content

        E=`openpkg_dev_execute --query`
        # E holds detection
        if [ ".${OPENPKG_EXEC}" = . ]; then
            if [ ".$E" = . ]; then
                # nothing forced, nothing detected; keep nothing; print nothing
                :
            else
                # nothing forced, something detected; keep detected; print detected normal
                EX="E=$E"
                ED="E=$E"
            fi
        else
            if [ ".$E" = . ]; then
                # forced, nothing detected; take forced; print forced black (aka trailing dot)
                E="${OPENPKG_EXEC}"
                EX="[1mE=$E[m"
                ED="E=$E."
            else
                if [ ".${OPENPKG_EXEC}" = ".$E" ]; then
                    # forced; detected and matched; keep; print black (aka trailing dot)
                    EX="[1mE=$E[m"
                    ED="E=$E."
                else
                    # forced; detected and mismatch; keep forced; print forced red (aka trailing exclamation mark)
                    EX="[1mE=${OPENPKG_EXEC}[m"
                    ED="E=${OPENPKG_EXEC}!"
                    E="${OPENPKG_EXEC}"
                fi
            fi
        fi
        # E has ultimate content

        T=`openpkg_dev_tools --query`
        # T holds detection
        if [ ".${OPENPKG_TRUN}" = . ]; then
            if [ ".$T" = . ]; then
                # nothing forced, nothing detected; keep nothing; print nothing
                :
            else
                # nothing forced, something detected; keep detected; print detected normal
                TX="T=$T"
                TD="T=$T"
            fi
        else
            if [ ".$T" = . ]; then
                # forced, nothing detected; take forced; print forced black (aka trailing dot)
                T="${OPENPKG_TRUN}"
                TX="[1mT=$T[m"
                TD="T=$T."
            else
                if [ ".${OPENPKG_TRUN}" = ".$T" ]; then
                    # forced; detected and matched; keep; print black (aka trailing dot)
                    TX="[1mT=$T[m"
                    TD="T=$T."
                else
                    # forced; detected and mismatch; keep forced; print forced red (aka trailing exclamation mark)
                    TX="[1mT=${OPENPKG_TRUN}[m"
                    TD="T=${OPENPKG_TRUN}!"
                    T="${OPENPKG_TRUN}"
                fi
            fi
        fi
        # T has ultimate content

        S=`openpkg_dev_srcdir --query`
        # S holds detection
        if [ ".${OPENPKG_SDIR}" = . ]; then
            if [ ".$S" = . ]; then
                # nothing forced, nothing detected; keep nothing; print nothing
                :
            else
                # nothing forced, something detected; keep detected; print detected normal
                SX="S=$S"
                SD="S=$S"
            fi
        else
            if [ ".$S" = . ]; then
                # forced, nothing detected; take forced; print forced black (aka trailing dot)
                S="${OPENPKG_SDIR}"
                SX="[1mS=$S[m"
                SD="S=$S."
            else
                if [ ".${OPENPKG_SDIR}" = ".$S" ]; then
                    # forced; detected and matched; keep; print black (aka trailing dot)
                    SX="[1mS=$S[m"
                    SD="S=$S."
                else
                    # forced; detected and mismatch; keep forced; print forced red (aka trailing exclamation mark)
                    SX="[1mS=${OPENPKG_SDIR}[m"
                    SD="S=${OPENPKG_SDIR}!"
                    S="${OPENPKG_SDIR}"
                fi
            fi
        fi
        # S has ultimate content

        IX=""
        if [ ".$PX" != . ]; then
            IX="${IX}${IX:+ }${PX}"
        fi
        if [ ".$BX" != . ]; then
            IX="${IX}${IX:+ }${BX}"
        fi
        if [ ".$EX" != . ]; then
            IX="${IX}${IX:+ }${EX}"
        fi
        if [ ".$TX" != . ]; then
            IX="${IX}${IX:+ }${TX}"
        fi
        if [ ".$SX" != . ]; then
            IX="${IX}${IX:+ }${SX}"
        fi
        if [ ".$IX" != . ]; then
            IX=" [${IX}]"
        fi

        ID=""
        if [ ".$PD" != . ]; then
            ID="${ID}${ID:+ }${PD}"
        fi
        if [ ".$BD" != . ]; then
            ID="${ID}${ID:+ }${BD}"
        fi
        if [ ".$ED" != . ]; then
            ID="${ID}${ID:+ }${ED}"
        fi
        if [ ".$TD" != . ]; then
            ID="${ID}${ID:+ }${TD}"
        fi
        if [ ".$SD" != . ]; then
            ID="${ID}${ID:+ }${SD}"
        fi
        if [ ".$ID" != . ]; then
            ID=" [${ID}]"
        fi

        if [ ".${TERM}" = .xterm ]; then
            PS1="${USER}@${HOSTNAME}:${PWD}${IX}\n${OPENPKG_MODE}\$ "
            echo -n "]0;openpkg dev ${USER}@${HOSTNAME}:${PWD}${ID}"
        else
            PS1="${USER}@${HOSTNAME}:${PWD}${ID}\n${OPENPKG_MODE}\$ "
        fi
    }
    alias opd="openpkg dev${OPENPKG_ARGS:+ }$OPENPKG_ARGS"
}

bashrcepilog () {
    #   sane environment
    umask 022

    #   prepare for a individual package
    [ ".$1" != . ] && openpkg_dev_package "$1"
    [ ".$2" != . ] && openpkg_dev_branch  "$2"
    [ ".$3" != . ] && openpkg_dev_execute "$3"
    [ ".$4" != . ] && openpkg_dev_tools   "$4"
    [ ".$5" != . ] && openpkg_dev_src     "$5"
    cd
}

bash () {
    #   honor optional parameters [[[spec]ctag]exec]; use "" to specify a empty parameter and use it's default
    [ ".$1" != . ] && OPENPKG_SPEC="$1"
    [ ".$2" != . ] && OPENPKG_CTAG="$2"
    [ ".$3" != . ] && OPENPKG_EXEC="$3"
    [ ".$4" != . ] && OPENPKG_TRUN="$4"
    [ ".$5" != . ] && OPENPKG_SDIR="$5"

    #   enter working directory
    builtin cd "${OPENPKG_WORK}" || die "cannot cd to ${OPENPKG_WORK}"
    P="${OPENPKG_SPEC}"

    #   canonify OPENPKG_CTAG
    canonifybranch "${OPENPKG_CTAG}"
    B="${RV}"

    #   canonify OPENPKG_EXEC
    matchinstance "${OPENPKG_EXEC}" exact
    E="${RV}"

    #   canonify OPENPKG_TRUN
    matchinstance "${OPENPKG_TRUN}" exact
    T="${RV}"

    #   we must have a E/bin/rpm somewhere
    if [ ".$E" = . ]; then
        matchinstance "${OPENPKG_CTAG}" any
        E="${RV}"
    fi

    #   we must have a T/bin/rpm somewhere
    if [ ".$T" = . ]; then
        matchinstance "${OPENPKG_CTAG}" any
        T="${RV}"
    fi

    #   create a unique .bashrc
                                     >.bashrc${GUID}
    dumpfunc -i bashrcprolog        >>.bashrc${GUID}
    dumpfunc    warn                >>.bashrc${GUID}
    dumpfunc    error               >>.bashrc${GUID}
    dumpfunc    makedir             >>.bashrc${GUID}
    dumpfunc    uwhich              >>.bashrc${GUID}
    dumpfunc    initmpx             >>.bashrc${GUID}
    dumpfunc    openpkgi            >>.bashrc${GUID}
    dumpfunc    openpkgE            >>.bashrc${GUID}
    dumpfunc    openpkgT            >>.bashrc${GUID}
    dumpfunc    openpkg             >>.bashrc${GUID}
    dumpfunc    rpmi                >>.bashrc${GUID}
    dumpfunc    rpmE                >>.bashrc${GUID}
    dumpfunc    rpmT                >>.bashrc${GUID}
    dumpfunc    rpm                 >>.bashrc${GUID}
    dumpfunc    rpm2cpioi           >>.bashrc${GUID}
    dumpfunc    rpm2cpioE           >>.bashrc${GUID}
    dumpfunc    rpm2cpioT           >>.bashrc${GUID}
    dumpfunc    rpm2cpio            >>.bashrc${GUID}
    dumpfunc    openpkg_dev_package >>.bashrc${GUID}
    dumpfunc    openpkg_dev_branch  >>.bashrc${GUID}
    dumpfunc    openpkg_dev_execute >>.bashrc${GUID}
    dumpfunc    openpkg_dev_tools   >>.bashrc${GUID}
    dumpfunc    openpkg_dev_srcdir  >>.bashrc${GUID}
    dumpfunc    _arg2sh             >>.bashrc${GUID}
    dumpfunc    _env2sh             >>.bashrc${GUID}
    dumpfunc    _root               >>.bashrc${GUID}
    dumpfunc    root                >>.bashrc${GUID}
    dumpfunc    cvs                 >>.bashrc${GUID}
    dumpfunc -i bashrcepilog        >>.bashrc${GUID}

    #   make sure temporary area exists
    if [ ! -d ${OPENPKG_TEMP} ]; then
        makedir 700 ${OPENPKG_TEMP} || die "cannot create ${OPENPKG_TEMP}"
    fi

    #   find reasonable bash
    matchinstance "" any
    if [ -f "${RV}/bin/bash" ]; then
        bash="${RV}/bin/bash"
    elif [ -f "${RV}/lib/openpkg/bash" ]; then
        bash="${RV}/lib/openpkg/bash"
    elif [ -f "`uwhich bash`" ]; then
        bash="`uwhich bash`"
    else
        die "no \"bash\" found in OpenPKG instance"
    fi

    #   enter environment
    echo "++ entering OpenPKG development shell (mode: [1m${OPENPKG_MODE}[m)"
    ${bash} --rcfile .bashrc${GUID}
    command rm .bashrc${GUID}
    echo "++ leaving OpenPKG development shell"

    #   remove workdir if it is completely empty
    #package -query workdir ${OPENPKG_SPEC} >/dev/null FIXME
    #rmdir $workdir 2>/dev/null                        FIXME
}

list () {
    if [ ".$1" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="$*"
    fi
    locationid=`rpmE --eval %{_rpmfilename} | sed -e 's;^.*}-;;' -e 's;\.rpm$;;'`
    for name in $all; do
        name=`echo $name | sed -e 's;\.src\.rpm$;;' -e 's;\.nosrc\.rpm$;;'` # convenience
        ls -l ${OPENPKG_WORK}/pkg/src/$name* | ${EGREP} "/$name[^-]*-[^-]+-[^-]+\.(no)?src.rpm$"
        ls -l ${OPENPKG_WORK}/pkg/bin/$name* | ${EGREP} "/$name[^-]*-[^-]+-[^-]+\.[^-]+-[^-]+-${locationid}.rpm$"
    done
}

peek () {
    if [ ".$1" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="$*"
    fi
    for name in $all; do
        name=`echo $name | sed -e 's;\.src\.rpm$;;' -e 's;\.nosrc\.rpm$;;'` # convenience
        if [ ! -f "${OPENPKG_WORK}/$S/${name}/${name}.spec" ]; then
            warn "${OPENPKG_WORK}/$S/${name}/${name}.spec not found"
            continue
        fi
        template=`rpmE --eval '%{_rpmfilename}'`
        rpmfile=`rpmE -q --qf "${template}" --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
        rpmdir=`rpmE --eval '%{_rpmdir}'`
        if [ ! -f "$rpmdir/$rpmfile" ]; then
            die "binary RPM not found: $rpmdir/$rpmfile"
        fi
        echo "++ peeking at file list of binary RPM: $rpmfile"
        filter="sed -e 's;^;\\ \\ \\ ;'"
        tmpfile="${OPENPKG_TEMP}/${name}${GUID}"
        for configfile in `rpmE -qplc $rpmdir/$rpmfile`; do
            filter="$filter -e \"s;^ \\\\(.* ${configfile}\\\\)\$;C\\\\1;\""
        done
        for docfile in `rpmE -qpld $rpmdir/$rpmfile`; do
            IFS="$OIFS"
            filter="$filter -e \"s;^ \\\\(.* ${docfile}\\\\)\$;D\\\\1;\""
        done
        rpmE -qplv $rpmdir/$rpmfile | eval $filter
    done
}

diff () {
    #   the name comes from the "cvs diff" command. It is actually a
    #   "dry" run of release() without altering anything.
    release -dry "$@"
}

release () {
    #   honor optional parameters [-dry] [-m "message"] [-x]
    dry=0
    extra=0
    msg=""
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-dry ) dry=1 ;;
            .-x   ) extra=1 ;;
            .-m   ) [ ".$2"    = . ] && die "-m option requires argument"
                    [ ".$msg" != . ] && die "-m can only be specified once"
                    shift
                    msg="$1" ;;
                * ) packages="${packages} $1"
        esac
        shift
    done

    #   examine if environment supports colorized diff
    if [ ".`uwhich cvs-diff-colorize`" != . ]; then
        cvs_diff_colorize="cvs-diff-colorize"
    else
        cvs_diff_colorize="cat"
    fi

    #   sanity check
    [ ".${OPENPKG_SPEC:-$P}" != ".$P" ] && die "release requires OPENPKG_SPEC (=${OPENPKG_SPEC}) and P (=$P) being in sync"
    [ ".${OPENPKG_CTAG:-$B}" != ".$B" ] && die "release requires OPENPKG_CTAG (=${OPENPKG_CTAG}) and B (=$B) being in sync"
    [ ".${OPENPKG_EXEC:-$E}" != ".$E" ] && die "release requires OPENPKG_EXEC (=${OPENPKG_EXEC}) and E (=$E) being in sync"

    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi

    #   make sure commit message from file uses absolute path
    if [ -f "${msg}" ]; then
        echo "${msg}" | ${EGREP} "^/" >/dev/null 2>&1 || msg="`pwd`/${msg}"
    fi

    #   determine whether CVS commit messages can be bundled
    bundle=0
    names=0
    for name in $all; do
        names=`expr $names + 1`
    done
    if [ ".${msg}" != . -a $names -gt 2 ]; then
        bundle=1
        bundle_files=""
        bundle_srpms=""
        bundle_wipes=""
        bundle_msg=""
        if [ -f "${msg}" ]; then
            bundle_msg="-F \"${msg}\""
        elif [ ".${msg}" = ".-" ]; then
            bundle_msg=""
        else
            bundle_msg="-m \"$msg\""
        fi
    fi

    for name in $all; do
        name=`echo $name | sed -e 's;\.src\.rpm$;;' -e 's;\.nosrc\.rpm$;;'` # convenience
        echo "++ processing package ${name}"
        if [ ! -f "${OPENPKG_WORK}/$S/${name}/${name}.spec" ]; then
            die "${OPENPKG_WORK}/$S/${name}/${name}.spec not found"
        fi
        template=`rpmE --eval '%{_rpmfilename}' | sed -e 's;\..*$;;'`
        rpmfile=`rpmE -q --qf "${template}" --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`

        echo "++ cleaning up sources and logs"
        rpmE --clean --nodeps ${OPENPKG_WORK}/$S/${name}/${name}.spec >/dev/null 2>&1 || true
        rm ${OPENPKG_WORK}/$S/${name}/.log-${name}@* >/dev/null 2>&1 || true

        # developer section
        if [ ".${OPENPKG_MODE}" = .developer ]; then

            # compatiblity FIXME
            echo "${OPENPKG_DIST}" | ${EGREP} "/current/SRC/?$" >/dev/null 2>&1
            if [ $? -eq 0 ]; then
                warn "deprecated suffix /current/src in OPENPKG_DIST=${OPENPKG_DIST}."
                warn "This might come from your ~/.openpkg/dev.rc file. Please fix."
                warn "Three second penalty :-)"
                sleep 3
                OPENPKG_DIST=`echo ${OPENPKG_DIST} | sed -e 's;/$;;' -e 's;/current/SRC$;;'`
            fi
            if [ ".${OPENPKG_WILD}" != . ]; then
                warn "deprecated use of OPENPKG_WILD. Do not use it."
                warn "Please list all instances in OPENPKG_INST (space delimited)."
                warn "This might come from your ~/.openpkg/dev.rc file. Please fix."
                warn "Three second penalty :-)"
                sleep 3
            fi

            echo "++ determine [no]src extension"
            ext=""
            [ -f "${OPENPKG_WORK}/pkg/src/${rpmfile}.src.rpm" ]   && ext="src.rpm"
            [ -f "${OPENPKG_WORK}/pkg/src/${rpmfile}.nosrc.rpm" ] && ext="nosrc.rpm"
            [ ".$ext" = . ] && die "${OPENPKG_WORK}/pkg/src/${rpmfile}.[no]src.rpm not found"

            _autorel

            if [ ".${OPENPKG_REPO}" = ".openpkg-cvs@cvs.openpkg.org:/v/openpkg/cvs" ]; then
                echo "++ checking release number"
                if [ ".$rpm_release_req" = . ]; then
                    die "cannot predict required release number for branch $B"
                fi
                if [ ".$rpm_release_new" != ".$rpm_release_req" ]; then
                    die "release number ${rpm_release_new} does not match expectation ${rpm_release_req}"
                fi

                echo "++ checking upload suffix"
                if [ ".$upload" = . ]; then
                    die "cannot determine upload suffix for branch $B"
                fi

                if [ ".${BRANCH}" = .HEAD ]; then
                    echo "++ linting ${name}.spec (package specification)"
                    openpkgT lint-spec ${OPENPKG_WORK}/$S/${name}/${name}.spec
                    if [ $? -gt 0 ]; then
                        die "unacceptable specification: ${OPENPKG_WORK}/$S/${name}/${name}.spec"
                    fi
                    if [ -f rc.${name} ]; then
                        echo "++ linting rc.${name} (run-command script)"
                        openpkgT lint-rc ${OPENPKG_WORK}/$S/${name}/rc.${name}
                        if [ $? -gt 0 ]; then
                            die "unacceptable run commands: ${OPENPKG_WORK}/$S/${name}/rc.${name}"
                        fi
                    fi
                    if [ -f fsl.${name} ]; then
                        echo "++ linting fsl.${name} (OSSP fsl configuration)"
                        openpkgT lint-fsl ${OPENPKG_WORK}/$S/${name}/fsl.${name}
                        if [ $? -gt 0 ]; then
                            die "unacceptable fsl configuration: ${OPENPKG_WORK}/$S/${name}/fsl.${name}"
                        fi
                    fi
                fi

                if [ ".${BRANCH}" = .HEAD -a ".$extra" = .1 ]; then
                    bin_template=`rpmE --eval '%{_rpmfilename}'`
                    bin_rpmfile=`rpmE -q --qf "${bin_template}" \
                                 --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
                    bin_rpmdir=`rpmE --eval '%{_rpmdir}'`
                    if [ -f "$bin_rpmdir/$bin_rpmfile" ]; then
                        echo "++ linting ${bin_rpmfile} (binary package)"
                        openpkgT lint-rpm --check=layout,attrib $bin_rpmdir/$bin_rpmfile
                        if [ $? -gt 0 ]; then
                            die "unacceptable specification: $bin_rpmdir/$bin_rpmfile"
                        fi
                    fi
                fi
            fi

            area=`echo ${OPENPKG_DIST}/${upload}/ | sed -e 's;openpkg-ftp@ftp.openpkg.org:/;OpenPKG ;'`
            uath=`echo ${OPENPKG_DIST} | sed -e 's;:.*$;;'`
            dist=`echo ${OPENPKG_DIST} | sed -e 's;^.*:;;'`
            exts="${ext}"
            if [ ".${name}" = .openpkg ]; then
                if [ ! -f "${OPENPKG_WORK}/pkg/src/${rpmfile}.src.sh" ]; then
                    die "required file ${OPENPKG_WORK}/pkg/src/${rpmfile}.src.sh missing"
                fi
                exts="${exts} src.sh"
            fi
            for ext in $exts; do
                echo "++ releasing ${rpmfile}.${ext} to distribution area ${area}"
                if [ ${bundle} -eq 1 ]; then
                    bundle_srpms="${bundle_srpms} ${OPENPKG_WORK}/pkg/src/${rpmfile}.${ext}"
                    if [ ".${rpmwipe}" != . ]; then
                        bundle_wipes="${bundle_wipes} ${OPENPKG_WORK}/pkg/src/${rpmwipe}.${ext}"
                    fi
                elif [ ${dry} -eq 0 ]; then
                    chmod 664 ${OPENPKG_WORK}/pkg/src/${rpmfile}.${ext}
                    if [ ".${rpmwipe}" != . ]; then
                        echo "rm ${dist}/${upload}/${rpmwipe}.${ext}" | sftp $uath
                    fi
                    scp -o "Compression no" -o "Cipher blowfish" "${OPENPKG_WORK}/pkg/src/${rpmfile}.${ext}" "${OPENPKG_DIST}/${upload}/"
                else
                    echo "chmod 664 ${OPENPKG_WORK}/pkg/src/${rpmfile}.${ext}"
                    if [ ".${rpmwipe}" != . ]; then
                        echo "echo \"rm ${dist}/${upload}/${rpmwipe}.${ext}\" | sftp $uath"
                    fi
                    echo "scp -o \"Compression no\" -o \"Cipher blowfish\" \"${OPENPKG_WORK}/pkg/src/${rpmfile}.${ext}\" \"${OPENPKG_DIST}/${upload}/\""
                fi
            done

            echo "++ determining commit message"
            if [ ".$msg" = . ]; then
                if [   ".${rpm_version_old}" = . \
                    -a ".${rpm_version_new}" != . \
                    -a ".${rpm_name}" != . \
                    -a ".${rpm_summary}" != . ]; then
                     msg="new package: ${rpm_name} ${rpm_version_new} (${rpm_summary})"
                elif [ ".${rpm_version_old}" != . \
                    -a ".${rpm_version_new}" != . \
                    -a ".${rpm_version_old}" != ".${rpm_version_new}" ]; then
                     msg="upgrading package: ${rpm_name} ${rpm_version_old} -> ${rpm_version_new}"
                elif [ ".${rpm_version_old}" != . \
                    -a ".${rpm_version_new}" != . \
                    -a ".${rpm_version_old}" = ".${rpm_version_new}" \
                    -a ".${rpm_release_old}" != ".${rpm_release_new}" ]; then
                     msg="modifying package: ${rpm_name}-${rpm_version_new} ${rpm_release_old} -> ${rpm_release_new}"
                elif [ ".${rpm_version_old}" != . \
                    -a ".${rpm_version_new}" != . \
                    -a ".${rpm_version_old}" = ".${rpm_version_new}" \
                    -a ".${rpm_release_old}" = ".${rpm_release_new}" ]; then
                     msg="modifying package: ${rpm_name}-${rpm_version_new} ${rpm_release_old} again"
                fi
                if [ ".${msg}" = . ]; then
                    die "unable to determine type of source changes. Please specify commit message"
                fi
            fi

            echo "++ determining commit file list"
            files="$S/${name}"

            echo "++ committing changes to CVS repository"
            builtin cd "${OPENPKG_WORK}" || die "cannot cd to ${OPENPKG_WORK}"
            if [ ".${msg}" != . -a -f "${msg}" ]; then
                if [ ${bundle} -eq 1 ]; then
                    bundle_files="${bundle_files} ${files}"
                elif [ ${dry} -eq 0 ]; then
                    cvs -d ${OPENPKG_REPO} commit -F "${msg}" $files
                else
                    echo cvs commit -F "${msg}" $files
                    cvs -d ${OPENPKG_REPO} diff $files | ${cvs_diff_colorize}
                fi
            elif [ ".${msg}" = ".-" ]; then
                if [ ${bundle} -eq 1 ]; then
                    bundle_files="${bundle_files} ${files}"
                elif [ ${dry} -eq 0 ]; then
                    cvs -d ${OPENPKG_REPO} commit $files
                else
                    echo cvs commit $files
                    cvs -d ${OPENPKG_REPO} diff $files | ${cvs_diff_colorize}
                fi
            else
                if [ ${bundle} -eq 1 ]; then
                    bundle_files="${bundle_files} ${files}"
                elif [ ${dry} -eq 0 ]; then
                    cvs -d ${OPENPKG_REPO} commit -m "${msg}" $files
                else
                    echo cvs commit -m "${msg}" $files
                    cvs -d ${OPENPKG_REPO} diff $files | ${cvs_diff_colorize}
                fi
            fi

        # contributor section
        else
            echo "++ determining contributor identification"
            shtoolE=`rpmE --eval '%{l_shtool}'`
            thisdate=`${shtoolE} echo -e '%D-%m-%Y'`
            username=`${shtoolE} echo -e '%u'`
            hostname=`${shtoolE} echo -e '%h%d'`
            realname=`(getent passwd; cat /etc/passwd; ypcat passwd; nidump passwd .) 2>/dev/null |\
                      grep "^${username}:" | ${AWK} -F: '{ print $5; }'`
            if [ ".$realname" = . ]; then
                realname="NN"
            fi

            echo "++ generating patch file (${OPENPKG_WORK}/${name}.diff)"
            ( echo "From: $realname <$username@$hostname>"
              echo "Date: $thisdate"
              if [ ".${msg}" = . ]; then
                  msg="unknown contribution"
              fi
              if [ -f "${msg}" ]; then
                  msg="`cat ${msg}`"
              fi
              echo "Subject: ${msg}"
              echo ""
              ( builtin cd ${OPENPKG_WORK}; cvs -d ${OPENPKG_REPO} diff -N -u3 $S/${name} )
            ) >${OPENPKG_WORK}/${name}.diff

            echo "++ uploading patch file to OpenPKG contribution area"
            curlE=`rpmE --eval '%{l_curl}'`
            if [ ${dry} -eq 0 ]; then
                $curlE --upload-file ${OPENPKG_WORK}/${name}.diff ${OPENPKG_DIST}/${name}.diff
            else
                echo $curlE --upload-file ${OPENPKG_WORK}/${name}.diff ${OPENPKG_DIST}/${name}.diff
                cat ${OPENPKG_WORK}/${name}.diff | ${cvs_diff_colorize}
            fi
        fi
    done

    #   lazy execution of bundled CVS commit
    if [ ${bundle} -eq 1 ]; then
        builtin cd "${OPENPKG_WORK}" || die "cannot cd to ${OPENPKG_WORK}"
        if [ ${dry} -eq 0 ]; then
            chmod 664 ${bundle_srpms}
            if [ ".${bundle_wipes}" != . ]; then
                echo "rm ${bundle_wipes}" | sftp $uath
            fi
            scp -o "Compression no" -o "Cipher blowfish" ${bundle_srpms} ${OPENPKG_DIST}/${upload}/
            eval "cvs -d ${OPENPKG_REPO} commit ${bundle_msg} ${bundle_files}"
        else
            echo chmod 664 ${bundle_srpms}
            if [ ".${bundle_wipes}" != . ]; then
                echo "echo \"rm ${bundle_wipes}\" | sftp $uath"
            fi
            echo "scp -o \"Compression no\" -o \"Cipher blowfish\" ${bundle_srpms} ${OPENPKG_DIST}/${upload}/"
            echo "cvs -d ${OPENPKG_REPO} commit ${bundle_msg} ${bundle_files}"
            cvs -d ${OPENPKG_REPO} diff ${bundle_files} | ${cvs_diff_colorize}
        fi
    fi
}

kill () {
    case ${OPENPKG_WORK} in
        /?* ) ;;
          * ) die "working area ${OPENPKG_WORK} seems to be strange" ;;
    esac
    if [ -d ${OPENPKG_WORK} ]; then
        echo "++ killing working area ${OPENPKG_WORK}"
        rm -rf ${OPENPKG_WORK} >/dev/null 2>&1 || true
    fi
}

clean () {
    cleanlogs=0
    cleantemp=0
    force=0
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
           .-l ) cleanlogs=1 ;;
           .-t ) cleantemp=1 ;;
           .-f ) force=1 ;;
           .-* ) die "unsupported option \"$1\"." ;;
             * ) packages="${packages} $1"
        esac
        shift
    done

    if [ $force -eq 1 -o $cleanlogs -eq 1 ]; then
        echo "++ cleaning up all logs in ${OPENPKG_WORK}/$S"
        rm ${OPENPKG_WORK}/$S/*/.log-*@* >/dev/null 2>&1 || true
    fi
    if [ $force -eq 1 -o $cleantemp -eq 1 ]; then
        case ${OPENPKG_TEMP} in
            /?* ) ;;
              * ) die "temporary area ${OPENPKG_TEMP} seems to be strange" ;;
        esac
        if [ -d ${OPENPKG_TEMP} ]; then
            echo "++ cleaning temporary area ${OPENPKG_TEMP}"
            rm -rf ${OPENPKG_TEMP}/* ${OPENPKG_TEMP}/.[a-z]* >/dev/null 2>&1 || true
        fi
    fi
    if [ $force -eq 0 -a $cleanlogs -eq 0 -a $cleantemp -eq 0 ]; then
        if [ ".${packages}" = . ]; then
            if [ ".$P" = . ]; then
                die "package cannot be identified. Neither parameter given nor \$P set."
            fi
            all="$P"
        else
            all="${packages}"
        fi
        for name in $all; do
            echo "++ cleaning up sources and logs for ${name}"
            rpmE --clean --nodeps ${OPENPKG_WORK}/$S/${name}/${name}.spec >/dev/null 2>&1 || true
            rm ${OPENPKG_WORK}/$S/${name}/.log-${name}@${HOSTNAME} >/dev/null 2>&1 || true
        done
    fi
}

new () {
    #   honor optional parameters [-dry]
    dry=0
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-dry ) dry=1 ;;
                * ) packages="${packages} $1"
        esac
        shift
    done

    if [ ".${packages}" = . ]; then
        die "package cannot be identified. Parameter is mandatory."
    else
        all="${packages}"
    fi

    for name in $all; do
        if [ -d "${OPENPKG_WORK}/$S/${name}" ]; then
            die "directory ${OPENPKG_WORK}/$S/${name} already exists -- remove first"
        fi

        echo "++ creating ${OPENPKG_RMOD}/${name}/ CVS area"
        if [ ${dry} -eq 0 ]; then
            makedir 755 ${OPENPKG_WORK}/$S/${name} || die "cannot create ${OPENPKG_WORK}/$S/${name}"
            if [ ".${OPENPKG_MODE}" = .developer ]; then
                ( builtin cd ${OPENPKG_WORK}/src && cvs -d ${OPENPKG_REPO} add ${name} )
            fi
        else
            echo "makedir 755 ${OPENPKG_WORK}/$S/${name}"
            if [ ".${OPENPKG_MODE}" = .developer ]; then
                echo "cd ${OPENPKG_WORK}/src && cvs -d ${OPENPKG_REPO} add ${name}"
            fi
        fi

        echo "++ creating ${OPENPKG_RMOD}/${name}/${name}.spec skeleton"
        year=`date '+%Y'`
        if [ ${dry} -eq 0 ]; then
            ( echo "##"
              echo "##  ${name}.spec -- OpenPKG RPM Package Specification"
              echo "##  Copyright (c) 2000-${year} OpenPKG Foundation e.V. <http://openpkg.net/>"
              echo "##  Copyright (c) 2000-${year} Ralf S. Engelschall <http://engelschall.com/>"
              echo "##"
              echo "##  Permission to use, copy, modify, and distribute this software for"
              echo "##  any purpose with or without fee is hereby granted, provided that"
              echo "##  the above copyright notice and this permission notice appear in all"
              echo "##  copies."
              echo "##"
              echo "##  THIS SOFTWARE IS PROVIDED \`\`AS IS'' AND ANY EXPRESSED OR IMPLIED"
              echo "##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF"
              echo "##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED."
              echo "##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR"
              echo "##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,"
              echo "##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT"
              echo "##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF"
              echo "##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND"
              echo "##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,"
              echo "##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT"
              echo "##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF"
              echo "##  SUCH DAMAGE."
              echo "##"
              echo ""
              echo "#   package information"
              echo "Name:         ${name}"
              echo "Summary:      Badly Packaged Program"
              echo "URL:          http://www.${name}.org/"
              echo "Vendor:       John Doe"
              echo "Packager:     OpenPKG"
              echo "Distribution: OpenPKG"
              echo "Class:        EVAL"
              echo "Group:        Unknown"
              echo "License:      DoePL"
              echo "Version:      7.42"
              echo "Release:      `date '+%Y%m%d'`"
              echo ""
              echo "#   list of sources"
              echo "Source0:      ftp://ftp.${name}.org/pub/${name}/${name}-%{version}.tar.gz"
              echo ""
              echo "#   build information"
              echo "Prefix:       %{l_prefix}"
              echo "BuildRoot:    %{l_buildroot}"
              echo "BuildPreReq:  OpenPKG, openpkg >= 20040130"
              echo "PreReq:       OpenPKG, openpkg >= 20040130"
              echo "AutoReq:      no"
              echo "AutoReqProv:  no"
              echo ""
              echo "%description"
              echo "    ${name} is the most badly packaged program in the OpenPKG world."
              echo ""
              echo "%track"
              echo "    prog ${name} = {"
              echo "        version   = %{version}"
              echo "        url       = ftp://ftp.${name}.org/pub/${name}/"
              echo "        regex     = ${name}-(__VER__)\\.tar\\.gz"
              echo "    }"
              echo ""
              echo "%prep"
              echo "    %setup -q"
              echo ""
              echo "%build"
              echo "    CC=\"%{l_cc}\" \\"
              echo "    CXX=\"%{l_cxx}\" \\"
              echo "    CFLAGS=\"%{l_cflags -O}\" \\"
              echo "    CXXFLAGS=\"%{l_cxxflags -O}\" \\"
              echo "    CPPFLAGS=\"%{l_cppflags}\" \\"
              echo "    LDFLAGS=\"%{l_ldflags}\" \\"
              echo "    ./configure \\"
              echo "        --prefix=%{l_prefix}"
              echo "    %{l_make} %{l_mflags -O}"
              echo ""
              echo "%install"
              echo "    rm -rf \$RPM_BUILD_ROOT"
              echo "    %{l_make} %{l_mflags} install AM_MAKEFLAGS=\"DESTDIR=\$RPM_BUILD_ROOT\""
              echo "    strip \$RPM_BUILD_ROOT%{l_prefix}/bin/* >/dev/null 2>&1 || true"
              echo "    %{l_rpmtool} files -v -ofiles -r\$RPM_BUILD_ROOT %{l_files_std}"
              echo ""
              echo "%files -f files"
              echo ""
              echo "%clean"
              echo "    rm -rf \$RPM_BUILD_ROOT"
              echo ""
            ) >${OPENPKG_WORK}/$S/${name}/${name}.spec
            if [ ".${OPENPKG_MODE}" = .developer ]; then
                ( builtin cd ${OPENPKG_WORK}/$S/${name} && cvs -d ${OPENPKG_REPO} add ${name}.spec )
            fi
        else
            echo "creating ${OPENPKG_WORK}/$S/${name}/${name}.spec"
            if [ ".${OPENPKG_MODE}" = .developer ]; then
                echo "cd ${OPENPKG_WORK}/$S/${name} && cvs -d ${OPENPKG_REPO} add ${name}.spec"
            fi
        fi
    done
}

erase () {
    force=0
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-f ) force=1 ;;
              * ) packages="${packages} $1"
        esac
        shift
    done
    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi
    erasethem=""
    for name in $all; do
        rpmE -q $name >/dev/null 2>&1
        if [ $? -ne 0 ]; then
            warn "nothing to erase for $name"
        else
            erasethem="$erasethem${erasethem:+ }$name"
        fi
    done
    if [ ".$erasethem" != . ]; then
        if [ ".$HOSTNAME" = . ]; then
            HOSTNAME=`hostname`
        fi
        initmpx $E
        if [ $force -ne 0 ]; then
            echo "++ erasing $erasethem forced"
            ssh -t -x "root@${HOSTNAME}" ${RPMCMD} -e "$erasethem" --nodeps
        else
            echo "++ erasing $erasethem"
            ssh -t -x "root@${HOSTNAME}" ${RPMCMD} -e "$erasethem"
        fi
        for name in $erasethem; do
            rpmE -q $name >/dev/null 2>&1
            if [ $? -eq 0 ]; then
                error "failed to erase $name"
            fi
        done
    fi
}

install () {
    force=0
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-f ) force=1 ;;
              * ) packages="${packages} $1"
        esac
        shift
    done
    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi
    for name in $all; do
        echo "++ installing ${name}"
        if [ ! -f "${OPENPKG_WORK}/$S/${name}/${name}.spec" ]; then
            warn "${OPENPKG_WORK}/$S/${name}/${name}.spec not found"
            continue
        fi
        package=`rpmE -q --qf "%{NAME}-%{VERSION}-%{RELEASE}" \
                 --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
        template=`rpmE --eval '%{_rpmfilename}'`
        rpmfile=`rpmE -q --qf "${template}" \
                 --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
        rpmdir=`rpmE --eval '%{_rpmdir}'`
        if [ ! -f "$rpmdir/$rpmfile" ]; then
            build ${name}
        fi
        if [ ".$HOSTNAME" = . ]; then
            HOSTNAME=`hostname`
        fi
        initmpx $E
        if [ $force -ne 0 ]; then
            echo "++ installing $package ($rpmfile) forced"
            ssh -t -x "root@${HOSTNAME}" ${RPMCMD} -Uvh "$rpmdir/$rpmfile" --force --nodeps
        else
            rpmE -q $package >/dev/null 2>&1 && continue
            echo "++ installing $package ($rpmfile)"
            ssh -t -x "root@${HOSTNAME}" ${RPMCMD} -Uvh "$rpmdir/$rpmfile"
        fi
        rpmE -q $package >/dev/null 2>&1
        if [ $? -ne 0 ]; then
            warn "failed to install $package"
        fi
        if [ ".${name}" != ".$P" ]; then
            echo "++ cleaning up sources and logs for ${name}"
            rpmE --clean --nodeps ${OPENPKG_WORK}/$S/${name}/${name}.spec >/dev/null 2>&1 || true
            rm ${OPENPKG_WORK}/$S/${name}/.log-${name}@${HOSTNAME} >/dev/null 2>&1 || true
        fi
    done
}

build () {
    force=0
    packages=""
    defall=""
    defpkg=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-D ) shift
                  opt="$1"
                  pkg=""; echo "${opt}" | ${EGREP} '::' >/dev/null 2>&1 && pkg="`echo ${opt} | sed -e 's;::.*$;;'`"
                  opt="`echo ${opt} | sed -e 's;^.*::;;'`"
                  echo "${opt}" | ${EGREP} '^with_' >/dev/null 2>&1 || opt="with_${opt}"
                  echo "${opt}" | ${EGREP} '='      >/dev/null 2>&1 || opt="${opt}=yes"
                  if [ ".${pkg}" = . ]; then
                      defall="${defall} ${opt}"
                  else
                      defpkg="${defpkg} ${pkg}::${opt}"
                  fi
                  ;;
            .-f ) force=1 ;;
              * ) packages="${packages} $1"
        esac
        shift
    done
    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi
    for name in $all; do
        opt="`echo ${defpkg} | tr ' ' '\n' | ${EGREP} ^$name:: | sed 's;^.*::;;' | tr '\n' ' '`"
        opt="`echo ${opt} $defall`"
        echo "++ building ${name} ${opt}"
        if [ ! -f "${OPENPKG_WORK}/$S/${name}/${name}.spec" ]; then
            warn "${OPENPKG_WORK}/$S/${name}/${name}.spec not found"
            continue
        fi

        if [ $force -eq 0 ]; then
            echo "++ checking release header having format suitable for automatic modification"
            ${EGREP} '^Release:  *((([0-9]\.)?[0-9]{8})|[0-9]+\.[0-9]+\.[0-9])$' <${OPENPKG_WORK}/$S/${name}/${name}.spec >/dev/null
            if [ $? -eq 0 -a ".${OPENPKG_REPO}" = ".openpkg-cvs@cvs.openpkg.org:/v/openpkg/cvs" ]; then
                echo "++ checking whether release header was already modified"
                (builtin cd ${OPENPKG_WORK}/$S/${name} && command cvs 2>&1 diff -u0 ${name}.spec) \
                | ${EGREP} '^[+-]Release: +' | wc -l | ${EGREP} '^ *2 *' >/dev/null
                if [ $? -ne 0 ]; then
                    echo "++ checking for local added, modified or removed files"
                    (builtin cd ${OPENPKG_WORK}/$S/${name} && command cvs 2>&1 stat) \
                    | ${EGREP} 'Status: *Locally' >/dev/null
                    if [ $? -eq 0 ]; then
                        _autorel
                        if [ ".$rpm_release_req" = . ]; then
                            die "cannot predict required release number for branch $B"
                        fi
                        if [ ".$rpm_release_new" != ".$rpm_release_req" ]; then
                            echo "++ modifying release number"
                            (   builtin cd ${OPENPKG_WORK}/$S/${name} \
                             && sed -e "s;^\(Release:  *\).*$;\1$rpm_release_req;" <${name}.spec >${name}.spec.n; \
                             mv ${name}.spec.n ${name}.spec)
                        fi
                    fi
                fi
            fi
        fi

        package=`rpmE -q --qf "%{NAME}-%{VERSION}-%{RELEASE}" \
                 --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
        template=`rpmE --eval '%{_rpmfilename}'`
        rpmfile=`rpmE -q --qf "${template}" \
                 --specfile ${OPENPKG_WORK}/$S/${name}/${name}.spec`
        rpmdir=`rpmE --eval '%{_rpmdir}'`
        if [ ! -f "$rpmdir/$rpmfile" -o ${OPENPKG_WORK}/$S/${name}/${name}.spec -nt "$rpmdir/$rpmfile" -o $force -ne 0 ]; then
            echo "++ building $package ($rpmfile)"
            opt="`echo ${opt} | tr ' ' '\n' | ${EGREP} . | sed -e 's;^;--define \";' -e 's;=; ;' -e 's;$;\";' | tr '\n' ' '`"
            (builtin cd ${OPENPKG_WORK}/$S/${name} && openpkgT fetch ${name}.spec)
            (builtin cd ${OPENPKG_WORK}/$S/${name} && eval "rpmE -ba ${opt} ${name}.spec" 2>&1 | tee .log-${name}@${HOSTNAME})
            if [ ! -f "$rpmdir/$rpmfile" ]; then
                warn "failed to build $package ($rpmfile)"
                continue
            fi
            if [ ".${name}" = .openpkg ]; then
                (builtin cd ${OPENPKG_WORK}/$S/${name} && eval "./openpkg.boot -s" 2>&1 | tee -a .log-${name}@${HOSTNAME})
            fi
        fi
    done
}

rpmbuildconvenience () {

    rpmbuild="$1"; shift

    force=0
    shortcircuit=""
    packages=""
    defall=""
    defpkg=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-D ) shift
                  opt="$1"
                  pkg=""; echo "${opt}" | ${EGREP} '::' >/dev/null 2>&1 && pkg="`echo ${opt} | sed -e 's;::.*$;;'`"
                  opt="`echo ${opt} | sed -e 's;^.*::;;'`"
                  echo "${opt}" | ${EGREP} '^with_' >/dev/null 2>&1 || opt="with_${opt}"
                  echo "${opt}" | ${EGREP} '='      >/dev/null 2>&1 || opt="${opt}=yes"
                  if [ ".${pkg}" = . ]; then
                      defall="${defall} ${opt}"
                  else
                      defpkg="${defpkg} ${pkg}::${opt}"
                  fi
                  ;;
            .-f ) force=1 ;;
            .-s ) shortcircuit="--short-circuit" ;;
            .-* ) die "unsupported option \"$1\"." ;;
              * ) packages="${packages} $1"
        esac
        shift
    done

    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi
    for name in $all; do
        opt="`echo ${defpkg} | tr ' ' '\n' | ${EGREP} ^$name:: | sed 's;^.*::;;' | tr '\n' ' '`"
        opt="`echo ${opt} $defall`"
        echo "++ processing ${name} ${opt}"
        if [ ! -f "${OPENPKG_WORK}/$S/${name}/${name}.spec" ]; then
            warn "${OPENPKG_WORK}/$S/${name}/${name}.spec not found"
            continue
        fi

        logfile="${OPENPKG_WORK}/$S/${name}/.log-${name}@${HOSTNAME}"
        if [ -f "$logfile" ]; then
            trim="TRUNC"
            if [ ".$shortcircuit" != . ]; then
                case ".$rpmbuild" in
                    .-bc ) trim="build" ;;
                    .-bi ) trim="install" ;;
                    .-bb ) trim="KEEP" ;;
                    .-bs ) trim="KEEP" ;;
                esac
            fi
            if [ ".$trim" = .TRUNC ]; then
                >$logfile
            elif [ ".$trim" != .KEEP ]; then
                sed -e "/^Executing(%${trim})/,\$d" <$logfile >$logfile.n && mv $logfile.n $logfile
            fi
        fi

        opt="`echo ${opt} | tr ' ' '\n' | ${EGREP} . | sed -e 's;^;--define \";' -e 's;=; ;' -e 's;$;\";' | tr '\n' ' '`"
        (builtin cd ${OPENPKG_WORK}/$S/${name} && eval "rpmE ${rpmbuild} ${shortcircuit} ${opt} ${name}.spec" 2>&1 | tee -a .log-${name}@${HOSTNAME})
        if [ ".$rpmbuild" = .-bs -a ".${name}" = .openpkg ]; then
            (builtin cd ${OPENPKG_WORK}/$S/${name} && eval "./openpkg.boot -s" 2>&1 | tee -a .log-${name}@${HOSTNAME})
        fi
    done
}

bbinary () {
    rpmbuildconvenience -bb "$@"
}

bcompile () {
    rpmbuildconvenience -bc "$@"
}

binstall () {
    rpmbuildconvenience -bi "$@"
}

bprep () {
    rpmbuildconvenience -bp "$@"
}

bsource () {
    rpmbuildconvenience -bs "$@"
}


logview () {
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
              * ) packages="${packages} $1"
        esac
        shift
    done
    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi
    for name in $all; do
        echo "++ logview ${name}"
        if [ -f "${OPENPKG_WORK}/$S/${name}/.log-${name}@${HOSTNAME}" ]; then
            cat "${OPENPKG_WORK}/$S/${name}/.log-${name}@${HOSTNAME}"
        else
            warn "log file for ${name} not found"
        fi
    done
}

search_for () {
    URL=`echo $2 | sed -e 's| *$||'`
    openpkg register --printstatus >/dev/null 2>&1 \
    && URL=`openpkg register --rewriteurls $URL`
    PKG=`$E/lib/openpkg/curl -l -s $URL |\
         sed -e 's|[^-a-zA-Z0-9_\.]||g' |\
         grep "^$1" | sort -r`
    if [ ".$PKG" != . ]; then
        for pkg in $PKG; do
            NAME=`echo "$pkg" | sed -e 's;\.src\.rpm$;;' -e 's;\.src\.sh$;;'`
            NAME=`echo "$NAME ........................................" | cut -c-35`
            DLOC=`echo "$pkg" | sed -e "s;^;$URL;"`
            echo "$NAME $DLOC"
        done
    fi
}

search () {
    #   we must have a E/bin/rpm somewhere
    if [ ".$E" = . ]; then
        matchinstance "${OPENPKG_CTAG}" any
        E="${RV}"
    fi

    if [ ".$1" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="$*"
    fi
    for i in $all; do
        search_for $i ftp://ftp.openpkg.org/current/SRC/
        search_for $i ftp://ftp.openpkg.org/stable/SRC/
        REL=`$E/lib/openpkg/curl -l -s ftp://ftp.openpkg.org/release/ |\
             sed -e 's;
;;' |\
             egrep '^[1-9][0-9]*\.[0-9]+$' | sort -rn`
        for r in $REL; do
            search_for $i ftp://ftp.openpkg.org/release/$r/UPD/
            search_for $i ftp://ftp.openpkg.org/release/$r/SRC/
            search_for $i ftp://ftp.openpkg.org/release/$r/SRC/PLUS/
        done
    done
}

lint () {
    force=0
    packages=""
    extra=0
    while [ ".$1" != . ]; do
        case ".$1" in
            .-f ) force=1 ;;
            .-x ) extra=1 ;;
              * ) packages="${packages} $1"
        esac
        shift
    done
    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi
    for name in $all; do
        echo "$name" | ${EGREP} "^[./]+" >/dev/null 2>&1
        if [ $? -ne 0 ]; then
            echo "$name" | ${EGREP} "/" >/dev/null 2>&1
            if [ $? -eq 0 ]; then
                name="${OPENPKG_WORK}/$S/$name"
            else
                name=`echo $name | sed -e 's;\.spec$;;'` # convenience
                name="${OPENPKG_WORK}/$S/$name/$name.spec"
            fi
        fi
        if [ -d $name ]; then
            files="`find $name -type f -name '*.spec' -print | sort`"
        else
            files="$name"
        fi
        for file in $files; do
            name=`echo ${file} | sed -e 's;^.*/;;'`
            path=`echo ${file} | sed -e "s;/$name\$;;"`
            name=`echo ${name} | sed -e 's;\.spec$;;'`

            echo "++ linting $file (package specification)"
            openpkgT lint-spec $file
            if [ $? -gt 0 -a $force -ne 1 ]; then
                die "unacceptable specification"
            fi

            if [ -f ${path}/rc.${name} ]; then
                echo "++ linting ${path}/rc.${name} (run-command script)"
                openpkgT lint-rc ${path}/rc.${name}
                if [ $? -gt 0 -a $force -ne 1 ]; then
                    die "unacceptable run commands: ${path}/rc.${name}"
                fi
            fi

            if [ -f ${path}/fsl.${name} ]; then
                echo "++ linting ${path}/fsl.${name} (OSSP fsl configuration)"
                openpkgT lint-fsl ${path}/fsl.${name}
                if [ $? -gt 0 -a $force -ne 1 ]; then
                    die "unacceptable fsl configuration: ${path}/fsl.${name}"
                fi
            fi

            if [ ".$extra" = .1 ]; then
                bin_template=`rpmE --eval '%{_rpmfilename}'`
                bin_rpmfile=`rpmE -q --qf "${bin_template}" --specfile $file`
                bin_rpmdir=`HOME=${OPENPKG_WORK} rpmE --eval '%{_rpmdir}'`
                if [ -f "$bin_rpmdir/$bin_rpmfile" ]; then
                    echo "++ linting ${bin_rpmdir}/${bin_rpmfile} (binary package)"
                    echo openpkgT lint-rpm --check=layout,attrib $bin_rpmdir/$bin_rpmfile
                    openpkgT lint-rpm --check=layout,attrib $bin_rpmdir/$bin_rpmfile
                    if [ $? -gt 0 -a $force -ne 1 ]; then
                        die "unacceptable specification: $bin_rpmdir/$bin_rpmfile"
                    fi
                fi
            fi
        done
    done
}

testit () {
    # FIXME consider test() [shell function] vs. test(1) [builtin] vs. rpm [non-OpenPKG] --test
    rpmE --help | grep -- --test >/dev/null && echo "${RPMCMD:-rpm} does not support --test" # FIXME s/echo/die/
    die "testit not yet implemented. Developer: consider testit vs. test(1)"
}

track () {
    rpmT --help | grep -- --track >/dev/null || die "${RPMCMD:-rpm} does not support --track"

    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-* ) die "unsupported option \"$1\"." ;;
              * ) packages="${packages} $1"
        esac
        shift
    done

    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi
    for name in $all; do
        (builtin cd ${OPENPKG_WORK}/$S/${name} && eval "rpmT --track ${name}.spec")
    done
}

run () {
    force=0
    dry=0
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
            .-c   ) shift; cmd="$1" ;;
            .-f   ) force=1 ;;
            .-dry ) dry=1 ;;
              .-* ) die "unsupported option \"$1\"." ;;
                * ) packages="${packages} $1"
        esac
        shift
    done

    if [ ".${cmd}" = . ]; then
        die "command for execution missing"
    fi

    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi

    for name in $all; do
        echo "$name" | ${EGREP} "^[./]+" >/dev/null 2>&1
        if [ $? -ne 0 ]; then
            echo "$name" | ${EGREP} "/" >/dev/null 2>&1
            if [ $? -eq 0 ]; then
                name="${OPENPKG_WORK}/$S/$name"
            else
                name=`echo $name | sed -e 's;\.spec$;;'` # convenience
                name="${OPENPKG_WORK}/$S/$name/$name.spec"
            fi
        fi
        if [ -d $name ]; then
            files="`find $name -type f -name '*.spec' -print | sort`"
        else
            files="$name"
        fi
        for file in $files; do
            echo "++ Running command for package specification: $file"
            tmpfile="${OPENPKG_TEMP}/${GUID}"
            ${AWK} -vcmd="$cmd" -vfile="$file" <${file} >${tmpfile} '
                BEGIN { flag=0 }
                /^%prep$/ { flag=1 }
                /^$/ { if (flag == 1) print "    "cmd" #"file; flag=0 }
                /.*/ { print $0 }'
            if [ ${dry} -eq 0 ]; then
                rpmE -bp ${tmpfile} --nodeps
                rc=$?
                rpmE --clean ${tmpfile} --nodeps
            else
                sed <${tmpfile} -n -e '/^%prep$/,/^$/p'
                rc=$?
            fi
            command rm ${tmpfile}
            if [ $rc -gt 0 -a $force -ne 1 ]; then
                die "command in %prep section of ${tmpfile} returned error"
            fi
        done
    done
}

fetch () {
    cleana=0
    cleans=0
    force=0
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
           .-a ) cleana=1 ;;
           .-s ) cleans=1 ;;
           .-f ) force=1 ;;
           .-* ) die "unsupported option \"$1\"." ;;
             * ) packages="${packages} $1"
        esac
        shift
    done

    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi

    for name in $all; do
        echo "$name" | ${EGREP} "^[./]+" >/dev/null 2>&1
        if [ $? -ne 0 ]; then
            echo "$name" | ${EGREP} "/" >/dev/null 2>&1
            if [ $? -eq 0 ]; then
                name="${OPENPKG_WORK}/$S/$name"
            else
                name=`echo $name | sed -e 's;\.spec$;;'` # convenience
                name="${OPENPKG_WORK}/$S/$name/$name.spec"
            fi
        fi
        if [ -d $name ]; then
            files="`find $name -type f -name '*.spec' -print | sort`"
        else
            files="$name"
        fi
        for file in $files; do
            if [ $cleana -ne 0 ]; then
                spec=`echo ${file} | sed -e 's;^.*/;;' -e 's;\.spec$;;'`
                echo "++ Cleaning out all dst files for $spec"
                for s in `builtin cd ${OPENPKG_WORK}/dst/${spec}/ && echo *`; do
                    f="${OPENPKG_WORK}/dst/${spec}/${s}"
                    [ -f "${f}" ] && command rm "${f}"
                    [ -e "${f}" -o -L "${f}" ] && die "Problem wiping out ${f}"
                done
            fi
            if [ $cleans -ne 0 ]; then
                spec=`echo ${file} | sed -e 's;^.*/;;' -e 's;\.spec$;;'`
                echo "++ Cleaning out selected dst files for $spec"
                for s in `builtin cd ${OPENPKG_WORK}/$S/${spec}/ && echo *`; do
                    f="${OPENPKG_WORK}/dst/${spec}/${s}"
                    [ -f "${f}" ] && command rm "${f}"
                    [ -e "${f}" -o -L "${f}" ] && die "Problem wiping out ${f}"
                done
            fi
            echo "++ Fetching files for package specification: $file"
            openpkgT fetch ${file}
            if [ $? -gt 0 -a $force -ne 1 ]; then
                die "fetching sources for ${file} failed"
            fi
        done
    done
}

vim () {
    force=0
    packages=""
    while [ ".$1" != . ]; do
        case ".$1" in
           .-f ) force=1 ;;
             * ) packages="${packages} $1"
        esac
        shift
    done

    if [ ".${packages}" = . ]; then
        if [ ".$P" = . ]; then
            die "package cannot be identified. Neither parameter given nor \$P set."
        fi
        all="$P"
    else
        all="${packages}"
    fi

    for name in $all; do
        echo "$name" | ${EGREP} "^[./]+" >/dev/null 2>&1
        if [ $? -ne 0 ]; then
            echo "$name" | ${EGREP} "/" >/dev/null 2>&1
            if [ $? -eq 0 ]; then
                name="${OPENPKG_WORK}/$S/$name"
            else
                name=`echo $name | sed -e 's;\.spec$;;'` # convenience
                name="${OPENPKG_WORK}/$S/$name/$name.spec"
            fi
        fi
        if [ -d $name ]; then
            files="`find $name -type f -name '*.spec' -print | sort`"
        else
            files="$name"
        fi
        for file in $files; do
            TERM=${TERM}-color command vim "${file}"
            if [ ".${OPENPKG_MODE}" = .developer -a ".${B}" = .HEAD -a ".${OPENPKG_REPO}" = ".openpkg-cvs@cvs.openpkg.org:/v/openpkg/cvs" -o $force -ne 0 ]; then
                openpkgT lint-spec ${file}
                if [ $? -gt 0 ]; then
                    die "unacceptable specification: ${file}"
                fi
            fi
        done
    done
}

query () {
    while [ ".$1" != . ]; do
        RV=""
        case ".$1" in
           .ma* ) shift; matchinstance $1 ;;
           .ca* ) shift; canonifybranch $1 ;;
              * ) RV="ERROR: unknow query \"$1\"" ;;
        esac
        echo "$RV"
        shift
    done
}

help () {
    if [ -d ${OPENPKG_WORK} ]; then
        out="tee ${OPENPKG_WORK}/00README"
    else
        out="cat"
    fi
    sed -e 's;^ *;;' <<EOF | $out

        OpenPKG Development
        ===================

        \$ openpkg dev [options] [cmds]

        options and environment variables
        --------------------------------------------------------------------------------------------------------
        -h|--help    ) Display brief usage message.

        --inst=... equal to setting OPENPKG_INST
        --work=... equal to setting OPENPKG_WORK
        --temp=... equal to setting OPENPKG_TEMP
        --mode=... equal to setting OPENPKG_MODE
        --repo=... equal to setting OPENPKG_REPO
        --dist=... equal to setting OPENPKG_DIST
        --spec=... equal to setting OPENPKG_SPEC or running cmd "package"
        --ctag=... equal to setting OPENPKG_CTAG or running cmd "branch"
        --exec=... equal to setting OPENPKG_EXEC or running cmd "execute"
        --trun=... equal to setting OPENPKG_TRUN or running cmd "tools"
        --sdir=... equal to setting OPENPKG_SDIR or running cmd "srcdir"

        cmds and parameters
        --------------------------------------------------------------------------------------------------------
        \$ openpkg dev bash    # [[[spec]ctag]exec]               enter development environment
        \$ openpkg dev bbinary # [[spec]...] [-f][-s][-D with...] rpmE -bb [--short-circuit] spec (force, opt)
        \$ openpkg dev bcompile# [[spec]...] [-f][-s][-D with...] rpmE -bc [--short-circuit] spec (force, opt)
        \$ openpkg dev binstall# [[spec]...] [-f][-s][-D with...] rpmE -bi [--short-circuit] spec (force, opt)
        \$ openpkg dev bprep   # [[spec]...] [-f]    [-D with...] rpmE -bp spec (force, opt)
        \$ openpkg dev branch  # [ctag]                           switch to a different branch
        \$ openpkg dev bsource # [[spec]...] [-f][-s][-D with...] rpmE -bs [--short-circuit] spec (force, opt)
        \$ openpkg dev build   # [[spec]...] [-f][-f][-D with...] build a package (force, option)
        \$ openpkg dev clean   # [-l][-t][-f] [[spec]...]         clean logs and/or temp (force all)
        \$ openpkg dev diff    #                                  alias for "release -dry"
        \$ openpkg dev erase   # [[spec]...] [-f]                 erase a package (forcibly)
        \$ openpkg dev execute # [exec]                           execute command on instance closest to branch
        \$ openpkg dev fetch   # [[spec]...] [-s][-a]             fetch sources; clean selected or all dst files
        \$ openpkg dev help    #                                  display this command summary and update 00README
        \$ openpkg dev install # [[spec]...] [-f]                 install a package (forcibly), build if required
        \$ openpkg dev kill    #                                  kill the environment
        \$ openpkg dev lint    # [spec|file|path]...] [-f]        lint package specification
        \$ openpkg dev list    # [[spec]...]                      list source and binary packages
        \$ openpkg dev logview # [[spec]...]                      view build logs
        \$ openpkg dev new     # spec... [-dry]                   create new package from scratch
        \$ openpkg dev package # [spec]                           switch to a different package
        \$ openpkg dev peek    # [[spec]...]                      peek at file list of binary package
        \$ openpkg dev query   # matchinstance|canonifybranch     query script internals
        \$ openpkg dev release # [[spec]...] [-dry] [-m "msg"]    release package changes to master server
        \$ openpkg dev run     # [[spec]...] [-dry] -c "cmd" [-f] append cmd to %prep for spec and run rpm -bp
        \$ openpkg dev search  # [[spec]...]                      search a package on master server
        \$ openpkg dev setup   # [[[spec]ctag]exec]               initial setup of openpkg dev working environment
        \$ openpkg dev srcdir  # [dir]                            use a different source directory
        \$ openpkg dev test    # [[spec]...]                      run functional test (reserved for future use)
        \$ openpkg dev tools   # [path]                           tools to run (default to instance internals)
        \$ openpkg dev track   # [[spec]...]                      track version
        \$ openpkg dev update  # [[spec]...] [-f][-j]             update specs (full) (join to HEAD)
        \$ openpkg dev vim     # [[spec]...] [-f]                 spec edit and lint (if HEAD or forced)
EOF
}

#   ------------------------------------------------------------

#   when invoked with -h|--help option, artifically replace cmd with "help"
if [ $help -eq 1 ]; then
    set -- help
fi

#   check whether EGREP was found
if [ ".${EGREP}" = . ]; then
    error "$0:ERROR: no egrep(1) found that really supports required regular expressions"
    exit 1
fi

#   check whether AWK was found
if [ ".${AWK}" = . ]; then
    error "$0:ERROR: no awk(1) found that supports the required \"-v var=val\" option"
    exit 1
fi

#   when invoked without any parameter, artifically inject "bash" command as default
if [ $# -eq 0 ]; then
    set -- bash
fi

#   extract command parameter
cmd="$1"
shift

#   create a unique id for unique temporary files
if [ ".$HOSTNAME" = . ]; then
    HOSTNAME=`hostname`
fi
GUID=".`date +%Y%m%d%H%M%S`-${HOSTNAME}-$$"

#   make sure environment is set up
[ ".${OPENPKG_INST}" = . ] && die "OPENPKG_INST is not set";
[ ".${OPENPKG_WORK}" = . ] && die "OPENPKG_WORK is not set";
[ ".${OPENPKG_REPO}" = . ] && die "OPENPKG_REPO is not set";
[ ".${OPENPKG_DIST}" = . ] && die "OPENPKG_DIST is not set";

#   handle command
case ${cmd} in
    bash    | ba* ) cmd="bash"    ;;
    bbinary | bb* ) cmd="bbinary" ;;
    bcompile| bc* ) cmd="bcompile";;
    binstall| bi* ) cmd="binstall";;
    bprep   | bp* ) cmd="bprep"   ;;
    branch  | br* ) cmd="branch"  ;;
    bsource | bs* ) cmd="bsource" ;;
    build   | bu* ) cmd="build"   ;;
    clean   | cl* ) cmd="clean"   ;;
    diff    | di* ) cmd="diff"    ;;
    erase   | er* ) cmd="erase"   ;;
    execute | ex* ) cmd="execute" ;;
    fetch   | fe* ) cmd="fetch"   ;;
    help    | he* ) cmd="help"    ;;
    install | in* ) cmd="install" ;;
    kill          ) cmd="kill"    ;;
    list    | ls* ) cmd="list"    ;;
    lint    | li* ) cmd="lint"    ;;
    logview | lo* ) cmd="logview" ;;
    new     | ne* ) cmd="new"     ;;
    package | pa* ) cmd="package" ;;
    peek    | pe* ) cmd="peek"    ;;
    query   | qu* ) cmd="query"   ;;
    release | re* ) cmd="release" ;;
    run     | ru* ) cmd="run"     ;;
    setup         ) cmd="setup"   ;;
    search  | se* ) cmd="search"  ;;
    srcdir  | sr* ) cmd="srcdir"  ;;
    test    | te* ) cmd="test"    ;;
    tools   | to* ) cmd="tools"   ;;
    track   | tr* ) cmd="track"   ;;
    update  | up* ) cmd="update"  ;;
    vim     | vi* ) cmd="vim"     ;;
    *) die "unknown command \"${cmd}\"" ;;
esac
${cmd} "$@"
exit $?

##
##  MANUAL PAGE
##

=pod

=head1 NAME

B<openpkg dev> - OpenPKG Development Shell

=head1 SYNOPSIS

B<openpkg dev> [I<dev-opt> ...] [I<cmd> [I<cmd-opt> ...] [I<cmd-arg> ...]]

=head1 DESCRIPTION

B<openpkg dev> is the Development Shell of the OpenPKG Tool Chain. It
is used by developers to work conveniently with the OpenPKG packaging
sources from the OpenPKG CVS Repository.

=head1 GLOBAL OPTIONS

The following global command options I<opt> exist:

=over 4

=item B<-h>, B<--help>

Display brief usage message.

=item B<--inst=>I<dir-list>

Specify white-space separated list of directories where OpenPKG
instances can be found. If omitted the variable C<OPENPKG_INST> is
read from F<~/openpkg/dev.rc>, read from environment or defaults to
"C</openpkg-dev /openpkg-22 /openpkg-21 /openpkg-20>".

=item B<--work=>I<dir>

Specify a directory where all the Development takes place. If omitted
the variable C<OPENPKG_WORK> is read from F<~/openpkg/dev.rc>, read from
environment or defaults to "C<${HOME}/work/openpkg>".

=item B<--temp=>I<dir>

Specify a directory used to place temporary files during the various
build stages. If omitted the variable C<OPENPKG_TEMP> is read
from F<~/openpkg/dev.rc>, read from environment or defaults to
"C<${TMPDIR-/tmp}/${LOGNAME}/openpkg>".

=item B<--mode=>I<contributor|developer>

Specify the mode of operation. Pick from "C<contributor>" or
"C<developer>". The former can upload modified package specifications
and upload modified packges via FTP into the contribution area, the
latter requires direct access to the project master server via SSH. If
omitted the variable C<OPENPKG_MODE> is read from F<~/openpkg/dev.rc>,
read from environment or defaults to "C<contributor>".

=item B<--repo=>I<dir>

Specify access to the repository for uploading
modified package specifications. Contributors use
"C<:pserver:anonymous@cvs.openpkg.org:/v/openpkg/cvs>" and developers
use "C<openpkg-cvs@cvs.openpkg.org:/v/openpkg/cvs>". If omitted the
variable C<OPENPKG_REPO> is read from F<~/openpkg/dev.rc>, read from
environment or defaults to the contributor setting.

=item B<--dist=>I<dir>

Specify access to the repository for uploading modified packages.
Contributors use "C<ftp://ftp.openpkg.org/contrib/00UPLOAD>" and
developers use "C<openpkg-ftp@ftp.openpkg.org:/>". If omitted the
variable C<OPENPKG_DIST> is read from F<~/openpkg/dev.rc>, read from
environment or defaults to the contributor setting.

=item B<--spec=>I<name>

Specify the name of the package to work with. This also sets the C<$P>
variable within the Development Shell which can be used to shorten
command line input. If omitted the variable C<OPENPKG_SPEC> is read
from F<~/openpkg/dev.rc>, read from environment. The default is empty
and makes C<$P> being computed from the CWD. While working within the
Development Shell this setting can be changed at any time by running the
"package" command.

=item B<--ctag=>I<branch>

Specify the name of the CVS tag (branch) to work with. This also sets
the C<$B> variable within the Development Shell which can be used to
shorten command line input. If omitted the variable C<OPENPKG_CTAG> is
read from F<~/openpkg/dev.rc>, read from environment. The default is
empty and makes C<$B> being computed from CVS files below the CWD. While
working within the Development Shell this setting can be changed at any
time by running the "branch" command.

=item B<--exec=>I<dir>

Specify the prefix of the OpenPKG instance used to execute packaging
commands. This also sets the C<$E> variable within the Development
Shell which can be used to shorten command line input. If omitted the
variable C<OPENPKG_EXEC> is read from F<~/openpkg/dev.rc>, read from
environment. The default is empty and makes C<$E> being computed by
finding appropriate commands in the path of C<OPENPKG_INST>. While
working within the Development Shell this setting can be changed at any
time by running the "exec" command.

=item B<--trun=>I<dir>

Specify the prefix of the OpenPKG instance used to run tools commands.
This also sets the C<$T> variable within the Development Shell which
can be used to shorten command line input. If omitted the variable
C<OPENPKG_TRUN> is read from F<~/openpkg/dev.rc>, read from environment.
The default is empty and makes C<$T> being computed by finding
appropriate commands in the path of C<OPENPKG_INST>. While working
within the Development Shell this setting can be changed at any time by
running the "trun" command.

=item B<--sdir=>I<subdir>

Specify the subdirectory below C<OPENPKG_WORK> where packages can be
found. This also sets the C<$S> variable within the Development Shell
which can be used to shorten command line input. If omitted the variable
C<OPENPKG_SDIR> is read from F<~/openpkg/dev.rc>, read from environment.
The default is empty and makes C<$S> being computed from the CWD with
fallback to "src". While working within the Development Shell this
setting can be changed at any time by running the "srcdir" command.

=back

=head1 FILES

The following paths are used by B<openpkg dev>:

=over 4

=item F<~/.openpkg/dev.rc>

=item F<~/work/openpkg/>

=item F</openpkg-dev>

=back

=head1 SEE ALSO

bash(1), C<openpkg man rpm>.

=head1 AUTHOR

Thomas Lotterer E<lt>thl@openpkg.orgE<gt>

=cut

